import click
from rich import print
import astropy.units as u
import matplotlib.pyplot as plt
import os
import re

from cosmoglobe.plot import plot as cplot
from cosmoglobe.plot import trace as ctrace
from cosmoglobe.plot import gnom as cgnom


from .plottools import *


@click.group()
def commands_plotting():
    pass

# fmt: off
@commands_plotting.command()
@click.argument("input", nargs=-1)
@click.option("-sig", default = 0, help="Signal to be plotted 0 by default (0, 1, 2 is interprated as IQU)",)
@click.option("-comp", default = None, type= click.STRING,help="Component name for autosetting plottingparameters",)
@click.option("-freq", default = None, type= click.FLOAT,help="Frequency in GHz needed for scaling maps when using a model object input",)
@click.option("-ticks", default = None, help="Min and max value for data. If None, uses 97.5th percentile.",)
@click.option("-min", default = None, type= click.FLOAT,help="Min value of colorbar, overrides autodetector.",)
@click.option("-max", default = None, type= click.FLOAT,help="Max value of colorbar, overrides autodetector.",)
@click.option("-range", "rng",  default = None,  type= click.FLOAT, help="Color bar range")
@click.option("-nocbar",is_flag=True,help='Adds colorbar ("cb" in filename)',)
@click.option("-unit", default = None, type= click.STRING,help="Set unit (Under color bar), has LaTeX functionality. Ex. $\mu$",)
@click.option("-fwhm", default = 0.0,type= click.FLOAT,help="FWHM of smoothing, in arcmin.",)
@click.option("-nside", default = None, type= click.INT,help="nside for optional ud_grade.",)
@click.option("-sample", default = -1, type= click.INT,help="Select sample when inputing .h5 file",)
@click.option("-mask", default = None, type= click.STRING,help="Masks input with specified maskfile.",)
@click.option("-maskfill", default = None, type= click.STRING,help='Color to fill masked area. for example "gray". Transparent by default.',)
@click.option("-cmap", default = None, type= click.STRING,help="Chose colormap (ex. sunburst, planck, etc). Available are matplotlib and cmasher. Also qualitative plotly [ex. q-Plotly-4 (q for qualitative 4 for max color)]",)
@click.option("-norm", default = None, help='"Linear" is normal, "log" Normalizesusingsemi-logscale',)
@click.option("-remove_dip",is_flag=True,help="Fits a dipole to the map and removes it.",)
@click.option("-remove_mono",is_flag=True,help="Fits a monopole to the map and removes it.",)
@click.option("-title", default = None, type= click.STRING,help="Sets the full figure title. Has LaTeX functionaliity (ex. $A_{s}$.)",)
@click.option("-rlabel", default = None, type= click.STRING,help="Sets the upper right title. Has LaTeX functionaliity (ex. $A_{s}$.)",)
@click.option("-llabel", default = None, type= click.STRING,help="Sets the upper left title. Has LaTeX functionaliity (ex. $A_{s}$.)",)
@click.option("-width", default = None, help="Size in inches. Default is fraction=0.5",)
@click.option("-fraction", default = 0.5,type= click.FLOAT,help="Fraction of latex page width",)
@click.option("-xsize", default = 1000, type= click.INT,help="figuresize in px",)
@click.option("-darkmode",is_flag=True,help='Plots all outlines in white for dark bakgrounds ("dark" in filename)',)
#@click.option("-interactive",is_flag=True,help="Uses healpys interactive mode, zooming is possible etc.",)
@click.option("-rot", default = None, help="Describe the rotation to apply. In the form (lon, lat, psi) (unit: degrees) : the point at longitude *lon* and latitude *lat* will be at the center. An additional rotation of angle *psi* around this direction is applied.",)
@click.option("-coord", default = None, help="Either one of 'G', 'E' or 'C' to describe the coordinate system of the map,",)
@click.option("-nest",is_flag=True,help="If True, ordering scheme is NESTED.",)
@click.option("-flip", default = "astro" ,type= click.STRING,help="Defines the convention of projection : 'astro' or 'geo'",)
@click.option("-graticule",is_flag=True,help="Adds graticule",)
@click.option("-graticule_labels",is_flag=True,help="longitude and latitude labels",)
@click.option("-return_only_data",is_flag=True,help="Returns figure",)
@click.option("-projection_type", default = "mollweide" ,type= click.STRING,help='{"aitoff", "hammer", "lambert", "mollweide", "cart", "3d", "polar"} type of the plot',)
@click.option("-cb_orientation", default = "horizontal" ,type= click.STRING,help='{"horizontal", "vertical"} color bar orientation',)
@click.option("-xlabel", default = None, type= click.STRING,help="set x axis label",)
@click.option("-ylabel", default = None, type= click.STRING,help="set y axis label",)
@click.option("-longitude_grid_spacing", default = 60, type= click.FLOAT,help="set x axis grid spacing",)
@click.option("-latitude_grid_spacing", default = 30, type= click.FLOAT,help="set y axis grid spacing",)
@click.option("-override_plot_properties", default = None, help='Override the following plot proporties: "cbar_shrink", "cbar_pad", "cbar_label_pad", "figure_width": width, "figure_size_ratio": ratio.',)
@click.option("-xtick_label_color", default = "black" ,type= click.STRING,help="color of longitude ticks",)
@click.option("-ytick_label_color", default = "black" ,type= click.STRING,help="color of latitude ticks",)
@click.option("-graticule_color", default = None, type= click.STRING,help="color of latitude ticks",)
@click.option("-fontsize", default = None, help="Dictionary with fontsizes: 'xlabel', 'ylabel', 'title', 'xtick_label', 'ytick_label', 'cbar_label', 'cbar_tick_label'.",)
@click.option("-phi_convention", default = "counterclockwise" ,type= click.STRING,help='convention on x-axis (phi), "counterclockwise" (default), "clockwise", "symmetrical" (phi as it is truly given) if "flip" is "geo", "phi_convention" should be set to "clockwise".',)
@click.option("-custom_xtick_labels", default = None, help="override x-axis tick labels",)
@click.option("-custom_ytick_labels", default = None, help="override y-axis tick labels",)
@click.option("-white_background",is_flag=True,help="Sets the background to be white. (Transparent by default [recommended])",)
@click.option("-png",is_flag=True,help="Saves output as .png ().pdf by default)",)
@click.option("-outdir", default = None, type= click.Path(exists=True),help="Output directory for plot",)
@click.option("-outname", default = None, type= click.STRING,help="Output filename, overwrites autonaming",)
@click.option("-show",is_flag=True,help="Displays the figure in addition to saving",)
@click.option("-gif",is_flag=True,help="Convert inputs to gif",)
@click.option("-fps", default = 3, type= click.INT,help="Gif speed, 300 by default.",)
@click.option("-dpi", default = 300, type= click.INT,help="DPI of output file. 300 by default.",)
# fmt: on
def plot(input, **kwargs):
    """
    Invokes the plot function from the command line
    input: filename of fits or h5 file, optional plotting paramameters
    """
    kwargs["cbar"] = not kwargs.pop("nocbar")
    kwargs["freq"] = None if kwargs["freq"] is None else kwargs["freq"] * u.GHz
    kwargs["fwhm"] = None if kwargs["fwhm"] is None else kwargs["fwhm"] * u.arcmin

    # Pop click-specific arguments
    gif = kwargs.pop("gif")
    show = kwargs.pop("show")
    dpi = kwargs.pop("dpi")
    fps = kwargs.pop("fps")
    png = kwargs.pop("png")
    white_background = kwargs.pop("white_background")
    outdir = kwargs.pop("outdir")
    outname = kwargs.pop("outname")

    imgs = []  # Container for images used with gif
    for i, filename in enumerate(input):
        # Super hacky gif fix because "hold" is not implemented in newvisufunc
        if gif and i > 0:
            kwargs["return_only_data"] = True

        # Actually plot the stuff
        img, params = cplot(filename, **kwargs)

        # Super hacky gif fix because "hold" is not implemented in newvisufunc
        if gif:
            if i > 0:
                longitude, latitude, grid_map = img
                img = plt.pcolormesh(
                    longitude,
                    latitude,
                    grid_map,
                    vmin=params["ticks"][0],
                    vmax=params["ticks"][-1],
                    rasterized=True,
                    cmap=load_cmap(params["cmap"]),
                    shading="auto",
                )

            """
            # Show title for gifs
            sample = re.search('_k([0-9]*)', filename)
            if sample:
                sample = sample.group(1).lstrip("0")
            else:
                sample = filename
            """
        if show:
            plt.show()

        """
        Filename formatting and  outputting
        """
        *path, fn = os.path.split(filename)
        fn, ftype = os.path.splitext(fn)

        comp = kwargs["comp"]
        freq = kwargs["freq"]
        fwhm = kwargs["fwhm"]

        if ftype == ".h5":
            fn += "_" + comp

        fn = fn.replace("_IQU_", "_").replace("_I_", "_")
        if freq is not None:
            fn = fn.replace(comp, f"{comp}-{int(freq.value)}GHz")
        if outdir:
            path = outdir

        tags = []
        tags.append(f"{str(int(fwhm.value))}arcmin") if float(fwhm.value) > 0 else None
        tags.append("cb") if kwargs["cbar"] else None
        tags.append("masked") if kwargs["mask"] else None
        tags.append("nodip") if kwargs["remove_dip"] else None
        tags.append("dark") if kwargs["darkmode"] else None
        if params["cmap"] is None:
            params["cmap"] = "planck"
        tags.append(f'c-{params["cmap"]}')

        nside_tag = "_n" + str(int(params["nside"]))
        if nside_tag in fn:
            fn = fn.replace(nside_tag, "")

        if isinstance(kwargs["sig"], int):
            sig = ["I", "Q", "U"][kwargs["sig"]]
        fn = fn + f'_{sig}_w{str(int(params["width"]))}' + nside_tag

        for i in tags:
            fn += f"_{i}"

        filetype = "png" if png else "pdf"
        if outname:
            filetype = "png" if outname.endswith(".png") else "pdf"

        fn += f".{filetype}"
        if outname:
            fn = outname

        if outdir:
            path = [outdir]

        if gif:
            # Append images for animation
            imgs.append([img])
        else:
            tp = False if white_background else True
            filetype = "png" if png else "pdf"
            path = "." if path[0] == "" else path[0]
            print(f"[bold green]Outputting {fn}[/bold green] to {path}")
            plt.savefig(
                f"{path}/{fn}",
                bbox_inches="tight",
                pad_inches=0.02,
                transparent=tp,
                format=filetype,
                dpi=dpi,
            )

    if gif:
        import matplotlib.animation as animation

        ani = animation.ArtistAnimation(plt.gcf(), imgs, blit=True)
        fn = fn.replace(filetype, "gif")
        print(f"[bold green]Outputting {fn}[/bold green]")
        ani.save(fn, dpi=dpi, writer=animation.PillowWriter(fps=fps))

# fmt: off
@commands_plotting.command()
@click.argument("input", type=click.STRING)
@click.option("-lon", default=0, type=click.INT)
@click.option("-lat", default=0, type=click.INT)
@click.option( "-comp", default=None, type=click.STRING, help="Component name for autosetting plottingparameters",)
@click.option( "-sig", default=0, help="Which sky signal to plot",)
@click.option( "-freq", default=None, type=click.FLOAT, help="Frequency in GHz needed for scaling maps when using a model object input",)
@click.option( "-nside", default=None, type=click.INT, help="nside for optional ud_grade.",)
@click.option("-size", default=20, type=click.INT)
@click.option( "-sample", default=-1, type=click.INT, help="Select sample when inputing .h5 file",)
@click.option( "-min", "vmin", default=None, type=click.FLOAT, help="Min value of colorbar, overrides autodetector.",)
@click.option( "-max", "vmax", default=None, type=click.FLOAT, help="Max value of colorbar, overrides autodetector.",)
@click.option( "-ticks", default=None, help="Min and max value for data. If None, uses 97.5th percentile.",)
@click.option("-range", "rng", default=None, type=click.FLOAT, help="Color bar range")
@click.option( "-llabel", default=None, type=click.STRING, help="Set title, has LaTeX functionality. Ex. $\mu$",)
@click.option( "-rlabel", default=None, type=click.STRING, help="Set title, has LaTeX functionality. Ex. $\mu$",)
@click.option( "-unit", default=None, type=click.STRING, help="Set unit (Under color bar), has LaTeX functionality. Ex. $\mu$",)
@click.option( "-cmap", default=None, help="Choose different color map (string), such as Jet or planck",)
@click.option( "-graticule", is_flag=True, help="Add graticule",)
@click.option( "-norm", default=None, help='"Linear" is normal, "log" Normalizesusingsemi-logscale',)
@click.option( "-nocbar", is_flag=True, help="remove colorbar",)
@click.option( "-cbar_pad", default=0.04, type=click.FLOAT, help="Colorbar space",)
@click.option( "-cbar_shrink", default=0.7, type=click.FLOAT, help="Colorbar size",)
@click.option( "-fwhm", default=0.0, type=click.FLOAT, help="FWHM of smoothing, in arcmin.",)
@click.option( "-remove_dip", default=None, type=click.STRING, help="Fits a dipole to the map and removes it. Specify mask or type auto.",)
@click.option( "-remove_mono", default=None, type=click.STRING, help="Fits a monopole to the map and removes it.",)
@click.option( "-fontsize", default=None, type=click.INT, help="Fontsize",)
@click.option( "-figsize", default=None, help="Figure size",)
@click.option( "-darkmode", is_flag=True, help="Turn on darkmode",)
@click.option( "-sub", default=None, help="sub (ex. (2,1,2))",)
@click.option( "-hold", is_flag=True, help="Hold figure",)
@click.option( "-reuse_axes", is_flag=True, help="Reuse axes",)
@click.option( "-outdir", default=None, type=click.Path(exists=True), help="Output directory for plot",)
@click.option( "-outname", default=None, type=click.STRING, help="Output filename, overwrites autonaming",)
@click.option( "-show", is_flag=True, help="Displays the figure in addition to saving",)
@click.option( "-png", is_flag=True, help="Saves output as .png ().pdf by default)",)
@click.option( "-dpi", default=300, type=click.INT, help="DPI of output file. 300 by default.",)
# fmt: on
def gnom(
    input,
    **kwargs
):
    """
    Plots a fits file in gnomonic view.
    It is wrapper on the cosmoglobe "gnom" function.s
    """
    kwargs["cbar"] = not kwargs.pop("nocbar")
    kwargs["freq"] = None if kwargs["freq"] is None else kwargs["freq"] * u.GHz
    kwargs["fwhm"] = None if kwargs["fwhm"] is None else kwargs["fwhm"] * u.arcmin
    kwargs["figsize"] = (3,4) if kwargs["figsize"] is None else tuple(kwargs["figsize"])
    
    # Pop click-specific arguments
    show = kwargs.pop("show")
    dpi = kwargs.pop("dpi")
    png = kwargs.pop("png")
    outdir = kwargs.pop("outdir")
    outname = kwargs.pop("outname")

    img, params = cgnom(
        input,
        **kwargs
    )

    if show:
        plt.show()
    else:
        *path, fn = os.path.split(input)
        if outname:
            filetype = "png" if outname.endswith(".png") else "pdf"
        else:
            fn = os.path.splitext(fn)[0] + f'_gnomonic_{["I","Q","U"][kwargs["sig"]]}_{kwargs["lon"]}lon{kwargs["lat"]}lat_{kwargs["size"]}x{kwargs["size"]}deg_c-{params["cmap"]}'
            filetype = "png" if png else "pdf"

        if outname:
            fn = outname
        if outdir:
            path = [outdir]
        path = "." if path[0] == "" else path[0]
        fn = os.path.splitext(fn)[0] + f".{filetype}"
        print(f"[bold green]Outputting {fn}[/bold green] to {path}")
        plt.savefig(f"{path}/{fn}", bbox_inches="tight", pad_inches=0.02, format=filetype, dpi=dpi)

# fmt: off
@commands_plotting.command()
@click.argument( "input",)
@click.option( "-dataset", default=None, type=click.STRING, help="HDF dataset to be plotted",)
@click.option( "-sig", default=0, type=click.INT, help="Signal to be plotted 0 by default (0, 1, 2 is interprated as IQU)",)
@click.option( "-labels", default=None, help="List of label values for each input",)
@click.option( "-hideval", is_flag=True, help="Hide mean and standard deviation of samples",)
@click.option( "-burnin", default=0, type=click.INT, help="Burn in",)
@click.option( "-xlabel", default=None, type=click.STRING, help="set x axis label",)
@click.option( "-ylabel", default=None, type=click.STRING, help="set y axis label",)
@click.option( "-nbins", default=None, type=click.INT, help="Number of bins",)
@click.option( "-cmap", default="tab10", type=click.STRING, help="Color map (tab10 by default)",)
@click.option( "-figsize", default=None, help="figure size",)
@click.option( "-darkmode", is_flag=True, help="Turn on darkmode",)
@click.option( "-sub", default=None, help="sub (ex. (2,1,2))",)
@click.option( "-hold", is_flag=True, help="Hold figure",)
@click.option( "-reuse_axes", is_flag=True, help="Reuse axes",)
@click.option( "-outdir", default=None, type=click.Path(exists=True), help="Output directory for plot",)
@click.option( "-outname", default=None, type=click.STRING, help="Output filename, overwrites autonaming",)
@click.option( "-show", is_flag=True, help="Displays the figure in addition to saving",)
@click.option( "-png", is_flag=True, help="Saves output as .png ().pdf by default)",)
@click.option( "-dpi", default=300, type=click.INT, help="DPI of output file. 300 by default.",)
# fmt: on
def trace(
    input,
    **kwargs,
):
    """
    Creates a trace plot off data in a HDF file.
    This is a click wrapper the same function in cosmoglobe.
    """
    kwargs["cbar"] = not kwargs.pop("nocbar")
    kwargs["freq"] = None if kwargs["freq"] is None else kwargs["freq"] * u.GHz
    kwargs["fwhm"] = None if kwargs["fwhm"] is None else kwargs["fwhm"] * u.arcmin
    kwargs["figsize"] = (8,3) if kwargs["figsize"] is None else tuple(kwargs["figsize"])
    
    # Pop click-specific arguments
    show = kwargs.pop("show")
    dpi = kwargs.pop("dpi")
    png = kwargs.pop("png")
    outdir = kwargs.pop("outdir")
    outname = kwargs.pop("outname")

    kwargs["labels"] = kwargs["labels"].split(" ")
    kwargs["showval"] = not kwargs.pop("hideval")

    ctrace(
        input,
        **kwargs,
    )

    if show:
        plt.show()
    else:
        *path, fn = os.path.split(input)
        if outname:
            filetype = "png" if outname.endswith(".png") else "pdf"
        else:
            filetype = "png" if png else "pdf"

        if outname:
            fn = outname
        if outdir:
            path = [outdir]
        path = "." if path[0] == "" else path[0]
        fn = os.path.splitext(fn)[0] + f".{filetype}"
        print(f"[bold green]Outputting {fn}[/bold green] to {path}")
        plt.savefig(f"{path}/{fn}", bbox_inches="tight", pad_inches=0.02, format=filetype, dpi=dpi)



import pkg_resources
from typing import List, Literal, Optional, Union

from astropy.utils.data import download_file

from cosmoglobe.h5.chain import Chain
from cosmoglobe.plot import gnom, hist, plot, spec, standalone_colorbar, trace
from cosmoglobe.sky._units import K_CMB, K_RJ
from cosmoglobe.sky.model import SkyModel

try:
    __version__ = pkg_resources.get_distribution(__name__).version
except pkg_resources.DistributionNotFound:  # pragma: no cover
    ...

_COMPSEP_DATA_URL = "https://sdc.uio.no/vol/cosmoglobe-data/BeyondPlanck/compsep/"
_BP_DATA_URL = "https://sdc.uio.no/vol/cosmoglobe-data/BeyondPlanck/releases/v2/"

def get_test_chain(cache: bool = True) -> Chain:
    """Returns a minimal Cosmoglobe chain.

    This function will download the chain if it is not already cached.
    The chain is of size 1.6 GB.
    
    Parameters
    ----------
    cache
        Boolean for wether or not to cache the downloaded chain. Defaults to True.
    
    Returns
    -------
    Chain
        Initialized Chainfile object.
    """

    filename = _COMPSEP_DATA_URL + "compsep_chain.h5"
    file = download_file(filename, cache=cache)

    return Chain(file)

def sky_model_from_chain(
    chain: Union[str, Chain],
    nside: int,
    components: Optional[List[str]] = None,
    model: str = "BeyondPlanck",
    samples: Optional[Union[range, int, Literal["all"]]] = -1,
    burn_in: Optional[int] = None,
):
    """Returns a SkyModel initialized from the chain.

    Parameters
    ----------
    chain
        Path to a Cosmoglobe chainfile or a `Chain`.
    nside
        Model HEALPIX map resolution parameter.
    components
        List of components to include in the model.
    model
        String representing which Cosmoglobe model to use. Defaults to
        BeyondPlanck.
    samples
        The sample number for which to extract the model. If the input
        is 'all', then the model will be initialized with the average of all
        samples in the chain. Defaults to the last sample in the chain (-1).
    burn_in
        If samples is 'all', a burn_in sample can be provided for which all
        subsequent samples are used in the averaging.

    Returns
    -------
        A SkyModel initialized from the chain.

    Examples
    --------
    >>> from cosmoglobe.sky import model_from_chain
    >>> sky_model = model_from_chain("path/to/chainfile.h5", nside=256)
    >>> print(sky_model)
    SkyModel(
        version: BeyondPlanck
        nside: 256
        components(
            (ame): AME(freq_peak)
            (cmb): CMB()
            (dust): ThermalDust(beta, T)
            (ff): FreeFree(T_e)
            (radio): Radio(alpha)
            (synch): Synchrotron(beta)
        )
    )
    """

    return SkyModel.from_chain(
        chain=chain,
        nside=nside,
        components=components,
        model=model,
        samples=samples,
        burn_in=burn_in,
    )


def sky_model(nside: int, cache: bool = True) -> SkyModel:
    """Downloads and caches the Cosmoglobe Sky model. 
    
    The downloaded file is ~700 MB and contains the mean of all parameters
    from all BP production chains.

    Parameters
    ----------
    nside
        Healpix resolution of the model.

    cache
        Boolean for wether or not to cache the downloaded chain. Defaults to True.
    
    Returns
    -------
        Cosmoglobe `SkyModel`.
    """

    filename = _BP_DATA_URL + "BP_mean_v2.h5"
    file = download_file(filename, cache=cache, timeout=1000)

    return SkyModel.from_chain(
        chain=file,
        nside=nside,
    )

__all__ = [
    "sky_model", 
    "get_test_chain", 
    "sky_model_from_chain", 
    "plot", 
    "spec", 
    "hist", 
    "gnom", 
    "trace", 
    "standalone_colorbar",
    "K_RJ",
    "K_CMB",
]



import numba
import numpy as np
#######################
# HELPFUL TOOLS BELOW #
#######################


@numba.njit(cache=True, fastmath=True)  # Speeding up by a lot!
def unpack_alms(maps, lmax):
    #print("Unpacking alms")
    mmax = lmax
    nmaps = len(maps)
    # Nalms is length of target alms
    Nalms = int(mmax * (2 * lmax + 1 - mmax) / 2 + lmax + 1)
    alms = np.zeros((nmaps, Nalms), dtype=np.complex128)
    # Unpack alms as output by commander
    for sig in range(nmaps):
        i = 0
        for l in range(lmax+1):
            j_real = l ** 2 + l
            alms[sig, i] = complex(maps[sig, j_real], 0.0)
            i += 1

        for m in range(1, lmax + 1):
            for l in range(m, lmax + 1):
                j_real = l ** 2 + l + m
                j_comp = l ** 2 + l - m

                alms[sig, i] = complex(maps[sig, j_real], maps[sig, j_comp],) / np.sqrt(2.0)

                i += 1
    return alms

def alm2fits_tool(input, dataset, nside, lmax, fwhm, save=True,):
    """
    Function for converting alms in hdf file to fits
    """
    import h5py
    import healpy as hp

    try:
        sample = int(dataset.split("/")[0])
        print(f"Using sample {sample}")
    except:
        print(f"No sample specified, fetching last smample")
        with h5py.File(input, "r") as f:
            sample = str(len(f.keys()) - 2).zfill(6)
            dataset=sample+"/"+dataset
        print(f"Sample {sample} found, dataset now {dataset}")

    with h5py.File(input, "r") as f:
        alms = f[dataset][()]
        lmax_h5 = f[f"{dataset[:-3]}lmax"][()]  # Get lmax from h5

    if lmax:
        # Check if chosen lmax is compatible with data
        if lmax > lmax_h5:
            print(
                "lmax larger than data allows: ", lmax_h5,
            )
            print("Please chose a value smaller than this")
    else:
        # Set lmax to default value
        lmax = lmax_h5
    mmax = lmax

    alms_unpacked = unpack_alms(alms, lmax)  # Unpack alms
    # If not amp map, set spin 0.
    if "amp_alm" in dataset:
        pol = True
        if np.shape(alms_unpacked)[0]==1:
            pol = False
    else:
        pol = False
    

    print(f"Making map from alms, setting lmax={lmax}, pol={pol}")
    maps = hp.sphtfunc.alm2map(alms_unpacked, int(nside), lmax=int(lmax), mmax=int(mmax), fwhm=arcmin2rad(fwhm), pol=pol, pixwin=True,)
    outfile = dataset.replace("/", "_")
    outfile = outfile.replace("_alm", "")
    if save:
        outfile += f"_{str(int(fwhm))}arcmin" if fwhm > 0.0 else ""
        hp.write_map(outfile + f"_n{str(nside)}_lmax{lmax}.fits", maps, overwrite=True, dtype=None)
    return maps, nside, lmax, fwhm, outfile

def h5handler(input, dataset, min, max, maxchain, output, fwhm, nside, command, pixweight=None, zerospin=False, lowmem=False, notchain=False, remove_mono=False, coadd=False,):
    """
    Function for calculating mean and stddev of signals in hdf file
    """
    # Check if you want to output a map
    import h5py
    import healpy as hp
    from tqdm import tqdm

    if (lowmem and command == np.std): #need to compute mean first
        mean_data = h5handler(input, dataset, min, max, maxchain, output, fwhm,
            nside, np.mean, pixweight, zerospin, lowmem, remove_mono=remove_mono)

    print()
    if command: print("{:-^50}".format(f" {dataset} calculating {command.__name__} "))
    print("{:-^50}".format(f" nside {nside}, {fwhm} arcmin smoothing "))


    if coadd:
        if dataset[0].endswith("map"):
            type = "map"
        elif dataset[0].endswith("rms"):
            type = "map"
        elif dataset[0].endswith("alm"):
            type = "alm"
        elif dataset[0].endswith("sigma"):
            type = "sigma"
        else:
            type = "data"
    else:
        if dataset.endswith("map"):
            type = "map"
        elif dataset.endswith("rms"):
            type = "map"
        elif dataset.endswith("alm"):
            type = "alm"
        elif dataset.endswith("sigma"):
            type = "sigma"
        else:
            type = "data"

    if (lowmem):
        nsamp = 0 #track number of samples
        first_samp = True #flag for first sample
    else:
        dats = []

    use_pixweights = False if pixweight == None else True
    maxnone = True if max == None else False  # set length of keys for maxchains>1
    pol = True if zerospin == False else False  # treat maps as TQU maps (polarization)
    for c in range(1, maxchain + 1):
        filename = input.replace("c0001", "c" + str(c).zfill(4))
        with h5py.File(filename, "r") as f:

            if notchain:
                data = f[dataset][()]
                if data.shape[0] == 1:
                    # Make sure its interpreted as I by healpy
                    # For non-polarization data, (1,npix) is not accepted by healpy
                    data = data.ravel()
                dats.append(data)
                continue

            if maxnone:
                # If no max is specified, chose last sample
                max = len(f.keys()) - 2

            print("{:-^48}".format(f" Samples {min} to {max} in {filename}"))

            for sample in tqdm(range(min, max + 1), ncols=80):
                # Identify dataset
                # alm, map or (sigma_l, which is recognized as l)

                # Unless output is ".fits" or "map", don't convert alms to map.
                alm2map = True if output.endswith((".fits", "map")) else False

                # HDF dataset path formatting
                s = str(sample).zfill(6)

                # Sets tag with type
                if coadd:
                    if pol:
                        maps = np.zeros((len(dataset)-1, 3, hp.nside2npix(nside)))
                        rmss = np.zeros((len(dataset)-1, 3, hp.nside2npix(nside)))
                    else:
                        maps = np.zeros((len(dataset)-1, 1, hp.nside2npix(nside)))
                        rmss = np.zeros((len(dataset)-1, 1, hp.nside2npix(nside)))
                    for i, dset in enumerate(dataset[:-1]):
                        tag = f"{s}/{dset}"
                        maps[i,:] = f[tag][()]
                        rmss[i,:] = f[tag.replace('map', 'rms')][()]
                    rmss[rmss == 0] = np.inf
                    mu = np.zeros(maps[0].shape)
                    den = np.zeros(maps[0].shape)
                    for i in range(len(maps)):
                        mu += maps[i]/rmss[i]**2
                        den += 1/rmss[i]**2
                    mu = hp.ma(mu)
                    den = hp.ma(den)
                    if 'rms' in dataset[0]:
                        data = 1/den**0.5
                    else:
                        data = mu/den


                else:
                    tag = f"{s}/{dataset}"
                    #print(f"Reading c{str(c).zfill(4)} {tag}")

                    # Check if map is available, if not, use alms.
                    # If alms is already chosen, no problem
                    try:
                        data = f[tag][()]
                        if len(data[0]) == 0:
                            tag = f"{tag[:-3]}map"
                            print(f"WARNING! No {type} data found, switching to map.")
                            data = f[tag][()]
                            type = "map"
                    except:
                        print(f"Found no dataset called {dataset}")
                        print(f"Trying alms instead {tag}")
                        try:
                            # Use alms instead (This takes longer and is not preferred)
                            tag = f"{tag[:-3]}alm"
                            type = "alm"
                            data = f[tag][()]
                        except:
                            print("Dataset not found.")

                # If data is alm, unpack.
                if type == "alm":
                    lmax_h5 = f[f"{tag[:-3]}lmax"][()]
                    data = unpack_alms(data, lmax_h5)  # Unpack alms
                    
                if data.shape[0] == 1:
                    # Make sure its interprated as I by healpy
                    # For non-polarization data, (1,npix) is not accepted by healpy
                    data = data.ravel()

                # If data is alm and calculating std. Bin to map and smooth first.
                if type == "alm" and command == np.std and alm2map:
                    #print(f"#{sample} --- alm2map with {fwhm} arcmin, lmax {lmax_h5} ---")
                    data = hp.alm2map(data, nside=nside, lmax=lmax_h5, fwhm=arcmin2rad(fwhm), pixwin=True,pol=pol,)

                # If data is map, smooth first.
                elif type == "map" and fwhm > 0.0 and (command == np.std or command == np.cov):
                    #print(f"#{sample} --- Smoothing map ---")
                    if use_pixweights:
                        data = hp.sphtfunc.smoothing(data, fwhm=arcmin2rad(fwhm),pol=pol,use_pixel_weights=True,datapath=pixweight)
                    else: #use ring weights
                        data = hp.sphtfunc.smoothing(data, fwhm=arcmin2rad(fwhm),pol=pol,use_weights=True)

                # Removing monopole - test
                if remove_mono:
                    if pol:
                        data[0], mono = hp.remove_monopole(data[0], gal_cut=30, fitval=True)
                    else:
                        data, mono = hp.remove_monopole(data, gal_cut=30, fitval=True)

                if (lowmem):
                    if (first_samp):
                        first_samp=False
                        if (command==np.mean):
                            dats=data.copy()
                        elif (command==np.std):
                            dats=(mean_data - data)**2
                        else:
                            print('     Unknown command {command}. Exiting')
                            exit()
                    else:
                        if (command==np.mean):
                            dats=dats+data
                        elif (command==np.std):
                            dats=dats+(mean_data - data)**2
                    nsamp+=1
                else:
                    # Append sample to list
                    dats.append(data)

    if (lowmem):
        if (command == np.mean):
            outdata = dats/nsamp
        elif (command == np.std):
            outdata = np.sqrt(dats/nsamp)
    else:
        # Convert list to array
        dats = np.array(dats)
        # Calculate std or mean
        if (command == np.cov):
            N  = dats.shape[0]
            m1 = dats - dats.sum(axis=0, keepdims=1)/N
            outdata = np.einsum('ijk,imk->jmk', m1, m1)/N
        else:
            outdata = command(dats, axis=0)
            outdata = command(dats, axis=0) if command else dats
    # Smoothing afterwards when calculating mean
    if type == "alm" and command == np.mean and alm2map:
        print(f"# --- alm2map mean with {fwhm} arcmin, lmax {lmax_h5} ---")
        outdata = hp.alm2map(
            outdata, nside=nside, lmax=lmax_h5, fwhm=arcmin2rad(fwhm), pixwin=True, pol=pol
        )

    if type == "map" and fwhm > 0.0 and command == np.mean:
        print(f"--- Smoothing mean map with {fwhm} arcmin,---")
        if use_pixweights:
            outdata = hp.sphtfunc.smoothing(outdata, fwhm=arcmin2rad(fwhm),pol=pol,use_pixel_weights=True,datapath=pixweight)
        else: #use ring weights
            outdata = hp.sphtfunc.smoothing(outdata, fwhm=arcmin2rad(fwhm),pol=pol,use_weights=True)

    # Outputs fits map if output name is .fits
    if output.endswith(".fits"):
        hp.write_map(output, outdata, overwrite=True, dtype=None)
    elif output.endswith(".dat"):
        while np.ndim(outdata)>2: 
            if outdata.shape[-1]==4:
                tdata = outdata[:,0,0]
                outdata = outdata[:,:,3]
                outdata[:,0] = tdata
            else:
                outdata = outdata[:,:,0]

        np.savetxt(output, outdata)
    return outdata

def arcmin2rad(arcmin):
    return arcmin * (2 * np.pi) / 21600

def legend_positions(df, y, scaling):
    """
    Calculate position of labels to the right in plot... 
    """
    positions = {}
    for column in y:
        positions[column] = df[column].values[-1] - 0.005

    def push(dpush):
        """
        ...by puting them to the last y value and
        pushing until no overlap
        """
        collisions = 0
        for column1, value1 in positions.items():
            for column2, value2 in positions.items():
                if column1 != column2:
                    dist = abs(value1-value2)
                    if dist < scaling:# 0.075: #0.075: #0.023:
                        collisions += 1
                        if value1 < value2:
                            positions[column1] -= dpush
                            positions[column2] += dpush
                        else:
                            positions[column1] += dpush
                            positions[column2] -= dpush
                            return True
    dpush = .001
    pushings = 0
    while True:
        if pushings == 1000:
            dpush*=10
            pushings = 0
        pushed = push(dpush)
        if not pushed:
            break

        pushings+=1

    return positions

def cmb(nu, A):
    """
    CMB blackbody spectrum
    """
    h = 6.62607e-34 # Planck's konstant
    k_b  = 1.38065e-23 # Boltzmanns konstant
    Tcmb = 2.7255      # K CMB Temperature

    x = h*nu/(k_b*Tcmb)
    g = (np.exp(x)-1)**2/(x**2*np.exp(x))
    s_cmb = A/g
    return s_cmb

def sync(nu, As, alpha, nuref=0.408):
    """
    Synchrotron spectrum using template
    """
    print("nuref", nuref)
    #alpha = 1., As = 30 K (30*1e6 muK)
    nu_0 = nuref*1e9 # 408 MHz
    from pathlib import Path
    synch_template = Path(__file__).parent / "Synchrotron_template_GHz_extended.txt"
    fnu, f = np.loadtxt(synch_template, unpack=True)
    f = np.interp(nu, fnu*1e9, f)
    f0 = np.interp(nu_0, nu, f) # Value of s at nu_0
    s_s = As*(nu_0/nu)**2*f/f0
    return s_s

def ffEM(nu,EM,Te):
    """
    Freefree spectrum using emission measure
    """
    #EM = 1 cm-3pc, Te= 500 #K
    T4 = Te*1e-4
    nu9 = nu/1e9 #Hz
    g_ff = np.log(np.exp(5.960-np.sqrt(3)/np.pi*np.log(nu9*T4**(-3./2.)))+np.e)
    tau = 0.05468*Te**(-3./2.)*nu9**(-2)*EM*g_ff
    s_ff = 1e6*Te*(1-np.exp(-tau))
    return s_ff

def ff(nu,A,Te, nuref=40.):
    """
    Freefree spectrum
    """
    h = 6.62607e-34 # Planck's konstant
    k_b  = 1.38065e-23 # Boltzmanns konstant

    nu_ref = nuref*1e9
    S =     np.log(np.exp(5.960 - np.sqrt(3.0)/np.pi * np.log(    nu/1e9*(Te/1e4)**-1.5))+2.71828)
    S_ref = np.log(np.exp(5.960 - np.sqrt(3.0)/np.pi * np.log(nu_ref/1e9*(Te/1e4)**-1.5))+2.71828)
    s_ff = A*S/S_ref*np.exp(-h*(nu-nu_ref)/k_b/Te)*(nu/nu_ref)**-2
    return s_ff

def sdust(nu, Asd, nu_p, polfrac, fnu = None, f_ = None, nuref=22.,):
    """
    Spinning dust spectrum using spdust2
    """
    nuref = nuref*1e9 
    scale = 30./nu_p

    try:
        f = np.interp(scale*nu, fnu, f_)
        f0 = np.interp(scale*nuref, fnu, f_) # Value of s at nu_0
    except:
        from pathlib import Path
        ame_template = Path(__file__).parent / "spdust2_cnm.dat"
        fnu, f_ = np.loadtxt(ame_template, unpack=True)
        fnu *= 1e9
        f = np.interp(scale*nu, fnu, f_)
        f0 = np.interp(scale*nuref, fnu, f_) # Value of s at nu_0
        
    s_sd = polfrac*Asd*(nuref/nu)**2*f/f0
    return s_sd

def tdust(nu,Ad,betad,Td,nuref=545.):
    """
    Thermal dust modified blackbody spectrum.
    """
    h = 6.62607e-34 # Planck's konstant
    k_b  = 1.38065e-23 # Boltzmanns konstant
    nu0=nuref*1e9
    gamma = h/(k_b*Td)
    s_d=Ad*(nu/nu0)**(betad+1)*(np.exp(gamma*nu0)-1)/(np.exp(gamma*nu)-1)
    return s_d

def lf(nu,Alf,betalf,nuref=30e9):
    """
    low frequency component spectrum (power law)
    """
    return Alf*(nu/nuref)**(betalf)

def line(nu, A, freq, conversion=1.0):
    """
    Line emission spectrum
    """
    if isinstance(nu, np.ndarray):
        return np.where(np.isclose(nu, 1e9*freq), A*conversion, 0.0)
    else:
        if np.isclose(nu, 1e9*freq[0]):
            return A*conversion
        else:
            return 0.0

def rspectrum(nu, r, sig, scaling=1.0):
    """
    Calculates the CMB amplituded given a value of r and requested modes
    """
    import camb
    from camb import model, initialpower
    import healpy as hp
    #Set up a new set of parameters for CAMB
    pars = camb.CAMBparams()
    #This function sets up CosmoMC-like settings, with one massive neutrino and helium set using BBN consistency
    pars.set_cosmology(H0=67.5, ombh2=0.022, omch2=0.122, mnu=0.06, omk=0, tau=0.06)
    pars.InitPower.set_params(As=2e-9, ns=0.965, r=r)
    lmax=6000
    pars.set_for_lmax(lmax,  lens_potential_accuracy=0)
    pars.WantTensors = True
    results = camb.get_results(pars)
    powers = results.get_cmb_power_spectra(params=pars, lmax=lmax, CMB_unit='muK', raw_cl=True,)


    l = np.arange(2,lmax+1)

    if sig == "TT":
        cl = powers['unlensed_scalar']
        signal = 0
    elif sig == "EE":
        cl = powers['unlensed_scalar']
        signal = 1
    elif sig == "BB":
        cl = powers['tensor']
        signal = 2

    bl = hp.gauss_beam(40/(180/np.pi*60), lmax,pol=True)
    A = np.sqrt(sum( 4*np.pi * cl[2:,signal]*bl[2:,signal]**2/(2*l+1) ))
    return cmb(nu, A*scaling)

def fits_handler(input, min, max, minchain, maxchain, chdir, output, fwhm, nside, zerospin, drop_missing, pixweight, command, lowmem=False, fields=None, write=False, coadd=False, rms_maps=None):
    """
    Function for handling fits files.
    """
    # Check if you want to output a map
    import healpy as hp
    from tqdm import tqdm
    import os

    if coadd:
        input_list = input
        input = input_list[0]

    if maxchain is None:
        maxchain = minchain + 1

    if (not input.endswith(".fits")):
        print("Input file must be a '.fits'-file")
        exit()

    if (lowmem and command == np.std): #need to compute mean first
        mean_data = fits_handler(input, min, max, minchain, maxchain, chdir, output, fwhm, nside, zerospin, drop_missing, pixweight, lowmem, np.mean, fields, write=False, coadd=coadd,)

    if (minchain > maxchain):
        print('Minimum chain number larger that maximum chain number. Exiting')
        exit()
    aline=input.split('/')
    dataset=aline[-1]
    print()
    if command: print("{:-^50}".format(f" {dataset} calculating {command.__name__} "))
    if (nside == None):
        print("{:-^50}".format(f" {fwhm} arcmin smoothing "))
    else:
        print("{:-^50}".format(f" nside {nside}, {fwhm} arcmin smoothing "))

    type = 'map'

    if (not lowmem):
        dats = []

    nsamp = 0 #track number of samples
    first_samp = True #flag for first sample

    use_pixweights = False if pixweight == None else True
    maxnone = True if max == None else False  # set length of keys for maxchains>1
    pol = True if zerospin == False else False  # treat maps as TQU maps (polarization)
    for c in range(minchain, maxchain + 1):
        if (chdir==None):
            filename = input.replace("c0001", "c" + str(c).zfill(4))
        else:
            if maxchain > minchain + 1:
                filename = chdir+'_c%i/'%(c)+input
            else:
                filename = f'{chdir}/{input}'
        basefile = filename.split("k000001")
        if coadd:
            basefiles = []
            filenames = []
            for i in range(len(input_list)):
                if (chdir==None):
                    filenames.append(input_list[i].replace("c0001", "c" + str(c).zfill(4)))
                else:
                    if maxchain > minchain + 1:
                        filenames.append(chdir+'_c%i/'%(c)+input_list[i])
                    else:
                        filenames.append(f'{chdir}/{input_list[i]}')
                basefiles.append(filenames[i].split("k000001"))


        if maxnone:
            # If no max is specified, find last sample of chain
            # Assume residual file of convention res_label_c0001_k000234.fits, 
            # i.e. final numbers of file are sample number
            max_found = False
            siter=min
            while (not max_found):
                filename = basefile[0]+'k'+str(siter).zfill(6)+basefile[1]

                if (os.path.isfile(filename)):
                    siter += 1
                else:
                    max_found = True
                    max = siter - 1

        else:
            if (first_samp):
                for chiter in range(minchain,maxchain + 1):
                    if (chdir==None):
                        tempname = input.replace("c0001", "c" + str(c).zfill(4))
                    else:
                        tempname = chdir+'_c%i/'%(c)+input
                        temp = tempname.split("k000001")

                    for siter in range(min,max+1):
                        tempf = temp[0]+'k'+str(siter).zfill(6)+temp[1]

                        if (not os.path.isfile(tempf)):
                            print('chain %i, sample %i missing'%(c,siter))
                            print(tempf)
                            if (not drop_missing):
                                exit()


        print("{:-^48}".format(f" Samples {min} to {max} in {filename}"))

        for sample in tqdm(range(min, max + 1), ncols=80):
                # dataset sample formatting
                filename = basefile[0]+'k'+str(sample).zfill(6)+basefile[1]                
                if (first_samp):
                    # Check which fields the input maps have
                    if (not os.path.isfile(filename)):
                        if (not drop_missing):
                            exit()
                        else:
                            continue
                    
                    _, header = hp.fitsfunc.read_map(filename, h=True, dtype=None)
                    if fields!=None:
                        nfields = 0
                        for par in header:
                            if (par[0] == 'TFIELDS'):
                                nfields = par[1]
                                break
                        if (nfields == 0):
                            print('No fields/maps in input file')
                            exit()
                        elif (nfields == 1):
                            fields=(0)
                        elif (nfields == 2):
                            fields=(0,1)
                        elif (nfields == 3):
                            fields=(0,1,2)
                    #print('   Reading fields ',fields)

                    nest = False
                    for par in header:
                        if (par[0] == 'ORDERING'):
                            if (not par[1] == 'RING'):
                                nest = True
                            break

                    nest = False
                    for par in header:
                        if (par[0] == 'NSIDE'):
                            nside_map = par[1]
                            break


                    if (not nside == None):
                        if (nside > nside_map):
                            print('   Specified nside larger than that of the input maps')
                            print('   Not up-grading the maps')
                            print('')

                if (not os.path.isfile(filename)):
                    if (not drop_missing):
                        exit()
                    else:
                        continue

                if coadd:
                    # loop over the bands to weighted average over
                    # use rms maps
                    if pol:
                        maps = np.zeros((len(rms_maps), 3, hp.nside2npix(nside)))
                        rmss = np.zeros((len(rms_maps), 3, hp.nside2npix(nside)))
                    else:
                        maps = np.zeros((len(rms_maps), 1, hp.nside2npix(nside)))
                        rmss = np.zeros((len(rms_maps), 1, hp.nside2npix(nside)))
                    for i in range(len(filenames)-1):
                        filename = basefiles[i][0]+'k'+str(sample).zfill(6)+basefiles[i][1]         
                        maps[i,:] = hp.read_map(filename, nest=nest)
                        rmss[i,:] = hp.read_map(f'{chdir}/{rms_maps[i]}', nest=nest)
                    rmss[rmss == 0] = np.inf
                    mu = np.zeros(maps[0].shape)
                    den = np.zeros(maps[0].shape)
                    for i in range(len(maps)):
                        mu += maps[i]/rmss[i]**2
                        den += 1/rmss[i]**2
                    mu = hp.ma(mu)
                    den = hp.ma(den)
                    if 'rms' in dataset[0]:
                        data = 1/den**0.5
                    else:
                        data = mu/den
                else:
                    data = hp.fitsfunc.read_map(filename,field=fields,h=False, nest=nest, dtype=None)
                if (nest): #need to reorder to ring-ordering
                    data = hp.pixelfunc.reorder(data,n2r=True)


                # degrading if relevant
                if (not nside == None):
                    if (nside < nside_map):
                        data=hp.pixelfunc.ud_grade(data,nside) #ordering=ring by default

                if data.shape[0] == 1:
                    # Make sure its interprated as I by healpy
                    # For non-polarization data, (1,npix) is not accepted by healpy
                    data = data.ravel()

                # If smoothing applied and calculating stddev, smooth first.
                if fwhm > 0.0 and (command == np.std or command == np.cov):
                    #print(f"#{sample} --- Smoothing map ---")
                    if use_pixweights:
                        data = hp.sphtfunc.smoothing(data, fwhm=arcmin2rad(fwhm),pol=pol,use_pixel_weights=True,datapath=pixweight)
                    else: #use ring weights
                        data = hp.sphtfunc.smoothing(data, fwhm=arcmin2rad(fwhm),pol=pol,use_weights=True)
                    
                if (lowmem):
                    if (first_samp):
                        if (command==np.mean):
                            dats=data.copy()
                        elif (command==np.std):
                            dats=(mean_data - data)**2
                        else:
                            print('     Unknown command {command}. Exiting')
                            exit()
                    else:
                        if (command==np.mean):
                            dats=dats+data
                        elif (command==np.std):
                            dats=dats+(mean_data - data)**2
                    nsamp+=1
                else:
                    # Append sample to list
                    dats.append(data)
                first_samp=False

    if (lowmem):
        if (command == np.mean):
            outdata = dats/nsamp
        elif (command == np.std):
            outdata = np.sqrt(dats/nsamp)
    else:
        # Convert list to array
        dats = np.array(dats)
        # Calculate std or mean
        if (command == np.cov):
            N  = dats.shape[0]
            m1 = dats - dats.sum(axis=0, keepdims=1)/N
            outdata = np.einsum('ijk,imk->jmk', m1, m1)/N
        else:
            outdata = command(dats, axis=0)

    # Smoothing afterwards when calculating mean
    if fwhm > 0.0 and command == np.mean:
        print(f"--- Smoothing mean map with {fwhm} arcmin,---")
        if use_pixweights:
            outdata = hp.sphtfunc.smoothing(outdata, fwhm=arcmin2rad(fwhm),pol=pol,use_pixel_weights=True,datapath=pixweight)
        else: #use ring weights
            outdata = hp.sphtfunc.smoothing(outdata, fwhm=arcmin2rad(fwhm),pol=pol,use_weights=True)
    # Outputs fits map if output name is .fits
    if write:
        print("debug 3")
        if output.endswith(".fits"):
            hp.write_map(output, outdata, overwrite=True, dtype=None)
        elif output.endswith(".dat"):
            np.savetxt(output, outdata)
    else:
        return outdata




import time
import sys
import os
import numpy as np
import healpy as hp
from cosmoglobe.release.tools import *
import cosmoglobe as cg


def format_fits(chain, extname, types, units, nside, burnin, maxchain, polar, component, fwhm, nu_ref_t, nu_ref_p, procver, filename, bndctr, restfreq, bndwid, cmin=1, cmax=None, chdir=None, fields=None, scale=1., coadd=False):
    print()
    print("{:#^80}".format(""))
    print("{:#^80}".format(f" Formatting and outputting {filename} "))
    print("{:#^80}".format(""))
    
    if coadd:
        header = get_header(extname, types, units, nside, polar, component[0], fwhm, nu_ref_t, nu_ref_p, procver, filename, bndctr, restfreq, bndwid,)
    else:
        header = get_header(extname, types, units, nside, polar, component, fwhm, nu_ref_t, nu_ref_p, procver, filename, bndctr, restfreq, bndwid,)
    dset = get_data(chain, extname, component, burnin, maxchain, fwhm, nside, types, cmin, cmax, chdir, fields, scale, polar, coadd,)

    print(f"{procver}/{filename}", dset.shape)
    hp.write_map(f"{procver}/{filename}", dset, column_names=types, column_units=units, coord="G", overwrite=True, extra_header=header, dtype=None)


def get_data(chain, extname, component, burnin, maxchain, fwhm, nside, types, cmin, cmax, chdir, fields=None, scale=1.0, polar=True, coadd=False,):



    if extname.endswith("CMB"):
        # Mean data
        amp_mean = h5handler(input=chain, dataset="cmb/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.mean,)

        # stddev data
        amp_stddev = h5handler(input=chain, dataset="cmb/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.std,)

        # Masks
        mask1 = np.zeros((hp.nside2npix(nside)))

        mask2 = np.zeros((hp.nside2npix(nside)))

        dset = np.zeros((len(types), hp.nside2npix(nside)))
        dset[0] = amp_mean[0, :]
        dset[1] = amp_mean[1, :]
        dset[2] = amp_mean[2, :]
        
        dset[3] = amp_stddev[0, :]
        dset[4] = amp_stddev[1, :]
        dset[5] = amp_stddev[2, :]
        
        dset[6] = mask1
        dset[7] = mask2

    if extname.endswith("RESAMP-T"):
        # Mean data
        amp_mean = h5handler(input=chain, dataset="cmb/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.mean,)

        # stddev data
        amp_stddev = h5handler(input=chain, dataset="cmb/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.std,)

        dset = np.zeros((len(types), hp.nside2npix(nside)))
        dset[0] = amp_mean
        dset[1] = amp_stddev
    elif extname.endswith("RESAMP-P"):
        # Mean data
        amp_mean = h5handler(input=chain, dataset="cmb_lowl/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.mean,)

        # stddev data
        amp_stddev = h5handler(input=chain, dataset="cmb_lowl/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.std,)

        dset = np.zeros((len(types), hp.nside2npix(nside)))
        dset[0] = amp_mean[0,:]
        dset[1] = amp_mean[1,:]
        dset[2] = amp_stddev[0,:]
        dset[3] = amp_stddev[1,:]
    elif extname.endswith("SYNCHROTRON"):
        # Mean data
        amp_mean = h5handler(input=chain, dataset="synch/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.mean,)
        beta_mean = h5handler(input=chain, dataset="synch/beta_map", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=0.0, nside=nside, command=np.mean,)

        # stddev data
        amp_stddev = h5handler(input=chain, dataset="synch/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.std,)
        beta_stddev = h5handler(input=chain, dataset="synch/beta_map", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=0.0, nside=nside, command=np.std,)

        dset = np.zeros((len(types), hp.nside2npix(nside)))

        dset[0] = amp_mean[0, :]
        dset[1] = amp_mean[1, :]
        dset[2] = amp_mean[2, :]
        dset[3] = np.sqrt(amp_mean[1, :]**2 + amp_mean[2, :]**2)

        dset[4] = beta_mean[0, :]
        dset[5] = beta_mean[1, :]

        dset[6] = amp_stddev[0, :]
        dset[7] = amp_stddev[1, :]
        dset[8] = amp_stddev[2, :]
        dset[9] = np.sqrt(amp_stddev[1, :]**2 + amp_stddev[2, :]**2)

        dset[10] = beta_stddev[0, :]
        dset[11] = beta_stddev[1, :]

    elif extname.endswith("DUST"):
        # Mean data
        amp_mean = h5handler(input=chain, dataset="dust/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.mean,)
        beta_mean = h5handler(input=chain, dataset="dust/beta_map", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=0.0, nside=nside, command=np.mean,)
        assert amp_mean.shape == beta_mean.shape, f"SED parameters have different nside"
        T_mean = h5handler(input=chain, dataset="dust/T_map", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=0.0, nside=nside, command=np.mean,)

        # stddev data
        amp_stddev = h5handler(input=chain, dataset="dust/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.std,)
        beta_stddev = h5handler(input=chain, dataset="dust/beta_map", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=0.0, nside=nside, command=np.std,)
        T_stddev = h5handler(input=chain, dataset="dust/T_map", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=0.0, nside=nside, command=np.std,)



        dset = np.zeros((len(types), hp.nside2npix(nside)))

        if len(types) == 6:

            dset[0] = amp_mean

            dset[1] = beta_mean

            dset[2] = T_mean

            dset[3] = amp_stddev

            dset[4] = beta_stddev

            dset[5] = T_stddev
        else:

            dset[0] = amp_mean[0, :]
            dset[1] = amp_mean[1, :]
            dset[2] = amp_mean[2, :]
            dset[3] = np.sqrt(amp_mean[1, :]**2 + amp_mean[2, :]**2)

            dset[4] = beta_mean[0, :]
            dset[5] = beta_mean[1, :]

            dset[6] = T_mean[0, :]
            dset[7] = T_mean[1, :]

            dset[8] = amp_stddev[0, :]
            dset[9] = amp_stddev[1, :]
            dset[10] = amp_stddev[2, :]
            dset[11] = np.sqrt(amp_stddev[1, :]**2 + amp_stddev[2, :]**2)

            dset[12] = beta_stddev[0, :]
            dset[13] = beta_stddev[1, :]

            dset[14] = T_stddev[0, :]
            dset[15] = T_stddev[1, :]

    elif extname.endswith("FREE-FREE"):
        # Mean data
        amp_mean = h5handler(input=chain, dataset="ff/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.mean,)
        Te_mean = h5handler(input=chain, dataset="ff/Te_map", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=0.0, nside=nside, command=np.mean,)

        # stddev data
        amp_stddev = h5handler(input=chain, dataset="ff/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.std,)
        Te_stddev = h5handler(input=chain, dataset="ff/Te_map", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=0.0, nside=nside, command=np.std,)

        dset = np.zeros((len(types), hp.nside2npix(nside)))

        dset[0] = amp_mean
        dset[1] = Te_mean

        dset[2] = amp_stddev
        dset[3] = Te_stddev

    elif extname.endswith("AME"):
        # Mean/std amplitude 
        amp_mean = h5handler(input=chain, dataset="ame/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.mean,)
        amp_stddev = h5handler(input=chain, dataset="ame/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.std,)

        # Mean/std spectral parameters
        c = cg.Chain(chain)
        comp_type = c.parameters['ame']['type']
        if comp_type == 'exponential':
            sed_mean = h5handler(input=chain, dataset="ame/beta_map", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=0.0, nside=nside, command=np.mean,)
            sed_stddev = h5handler(input=chain, dataset="ame/beta_map", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=0.0, nside=nside, command=np.std,)
        elif comp_type == 'spindust2':
            sed_mean = h5handler(input=chain, dataset="ame/nu_p_map", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=0.0, nside=nside, command=np.mean,)
            sed_stddev = h5handler(input=chain, dataset="ame/nu_p_map", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=0.0, nside=nside, command=np.std,)
        else:
            print(f'Component type {comp_type} not supported')

        dset = np.zeros((len(types), hp.nside2npix(nside)))

        dset[0] = amp_mean
        dset[1] = sed_mean

        dset[2] = amp_stddev
        dset[3] = sed_stddev

    elif extname.endswith("CII-line"):
        # Mean/std amplitude 
        amp_mean = h5handler(input=chain, dataset="dust_cii/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.mean,)
        amp_stddev = h5handler(input=chain, dataset="dust_cii/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.std,)

        dset = np.zeros((len(types), hp.nside2npix(nside)))

        dset[0] = amp_mean
        dset[1] = amp_stddev

    elif extname.endswith("stars"):
        # Mean/std amplitude 
        # amp_mean = h5handler(input=chain, dataset="dust_cii/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.mean,)
        # amp_stddev = h5handler(input=chain, dataset="dust_cii/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.std,)

        amp_mean = fits_handler(input="stars_01a_c0001_k000001.fits", min=burnin, max=None, minchain=cmin, maxchain=cmax, chdir=chdir, output="map", fwhm=fwhm, nside=nside, drop_missing=True, pixweight=None, command=np.mean, lowmem=False, fields=fields, write=False, zerospin=False)
        amp_stddev = fits_handler(input="stars_01a_c0001_k000001.fits", min=burnin, max=None, minchain=cmin, maxchain=cmax, chdir=chdir, output="map", fwhm=fwhm, nside=nside, drop_missing=True, pixweight=None, command=np.std, lowmem=False, fields=fields, write=False, zerospin=False)
        #amp_mean = fits_handler(input="chisq_c0001_k000001.fits", min=burnin, max=None, minchain=cmin, maxchain=cmax, chdir=chdir, output="map", fwhm=fwhm, nside=nside, zerospin=False, drop_missing=True, pixweight=None, command=np.mean, lowmem=False, write=False)

        dset = np.zeros((len(types), hp.nside2npix(nside)))

        dset[0] = amp_mean
        dset[1] = amp_stddev

    elif extname.endswith("PAH"):
        # Mean/std amplitude 
        amp_mean = h5handler(input=chain, dataset="hotPAH/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.mean,)
        amp_stddev = h5handler(input=chain, dataset="hotPAH/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.std,)

        dset = np.zeros((len(types), hp.nside2npix(nside)))

        dset[0] = amp_mean
        dset[1] = amp_stddev
    elif extname.endswith("CO_tot"):
        # Mean/std amplitude 
        amp_mean = h5handler(input=chain, dataset="co_tot/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.mean,)
        amp_stddev = h5handler(input=chain, dataset="co_tot/amp_alm", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.std,)

        dset = np.zeros((len(types), hp.nside2npix(nside)))

        dset[0] = amp_mean
        dset[1] = amp_stddev

    elif extname.endswith("FREQMAP"):
        if polar:
            zerospin=False
        else:
            zerospin=True
        # Mean data
        if coadd:
            datasets = [f"tod/{comp}/map" for comp in component]
            amp_mean = h5handler(input=chain, dataset=datasets, min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.mean, zerospin=zerospin, coadd=coadd, )

            if polar:
                amp_covar = h5handler(input=chain, dataset=datasets, min=burnin, max=None, maxchain=maxchain, output="map", fwhm=120., nside=nside, command=np.cov, remove_mono=True, zerospin=zerospin, coadd=coadd,)
            else:
                amp_stddev = h5handler(input=chain, dataset=datasets, min=burnin, max=None, maxchain=maxchain, output="map", fwhm=120., nside=nside, command=np.std, remove_mono=True, zerospin=zerospin, coadd=coadd,)

            datasets = [f"tod/{comp}/rms" for comp in component]
            amp_rms  = h5handler(input=chain, dataset=datasets, min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.mean, zerospin=zerospin, coadd=coadd,)

        else:
            amp_mean = h5handler(input=chain, dataset=f"tod/{component}/map", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.mean, zerospin=zerospin,)
            amp_rms  = h5handler(input=chain, dataset=f"tod/{component}/rms", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.mean, zerospin=zerospin,)
            if polar:
                amp_covar = h5handler(input=chain, dataset=f"tod/{component}/map", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=120., nside=nside, command=np.cov, remove_mono=True, zerospin=zerospin,)
            else:
                amp_stddev = h5handler(input=chain, dataset=f"tod/{component}/map", min=burnin, max=None, maxchain=maxchain, output="map", fwhm=120., nside=nside, command=np.std, remove_mono=True, zerospin=zerospin,)

        # Masks

        dset = np.zeros((len(types), hp.nside2npix(nside)))


        if polar:
            dset[0] = amp_mean[0, :]
            dset[1] = amp_mean[1, :]
            dset[2] = amp_mean[2, :]

            dset[3] = amp_rms[0, :]**0.5
            dset[4] = amp_rms[1, :]**0.5
            dset[5] = amp_rms[2, :]**0.5
            dset[6] = amp_rms[3, :]

            dset[7] = amp_covar[0, 0, :]**0.5
            dset[8] = amp_covar[1, 1, :]**0.5
            dset[9] = amp_covar[2, 2, :]**0.5
            dset[10] = amp_covar[1, 2, :]
        else:
            dset[0] = amp_mean
            dset[1] = amp_rms
            dset[2] = amp_stddev


    elif extname.endswith("RES"):
        N = len(types)
        if polar:
            zerospin=False
        else:
            zerospin=True
        if coadd:
        #    datasets = [f"tod/{comp}/map" for comp in component]
        #    amp_mean = h5handler(input=chain, dataset=datasets, min=burnin, max=None, maxchain=maxchain, output="map", fwhm=fwhm, nside=nside, command=np.mean, zerospin=zerospin, coadd=coadd, )

        #    amp_stddev = h5handler(input=chain, dataset=datasets, min=burnin, max=None, maxchain=maxchain, output="map", fwhm=120., nside=nside, command=np.std, remove_mono=True, zerospin=zerospin, coadd=coadd,)
            # component is a list that looks like
            # 10a, 10b, 10

            inputs = [f"res_{c}_c0001_k000001.fits" for c in component]
            rms_maps = [f"tod_{c}_rms_c0001_k000001.fits" for c in component[:-1]]
            amp_mean = fits_handler(input=inputs, min=burnin, max=None, minchain=cmin, maxchain=cmax, chdir=chdir, output="map", fwhm=fwhm, nside=nside, zerospin=zerospin, drop_missing=True, pixweight=None, command=np.mean, lowmem=False, fields=fields, write=False, coadd=True, rms_maps=rms_maps)
            amp_stddev = fits_handler(input=inputs, min=burnin, max=None, minchain=cmin, maxchain=cmax, chdir=chdir, output="map", fwhm=fwhm, nside=nside, zerospin=zerospin, drop_missing=True, pixweight=None, command=np.std, lowmem=False, fields=fields, write=False, coadd=True, rms_maps=rms_maps)

        else:
            amp_mean = fits_handler(input=f"res_{component}_c0001_k000001.fits", min=burnin, max=None, minchain=cmin, maxchain=cmax, chdir=chdir, output="map", fwhm=fwhm, nside=nside, zerospin=zerospin, drop_missing=True, pixweight=None, command=np.mean, lowmem=False, fields=fields, write=False)
            amp_stddev = fits_handler(input=f"res_{component}_c0001_k000001.fits", min=burnin, max=None, minchain=cmin, maxchain=cmax, chdir=chdir, output="map", fwhm=fwhm, nside=nside, zerospin=zerospin, drop_missing=True, pixweight=None, command=np.std, lowmem=False, fields=fields, write=False)
        dset = np.zeros((N, hp.nside2npix(nside)))
        if len(fields)>1:
            dset[:N//2] = amp_mean[fields, :]*scale
            dset[N//2:] = amp_stddev[fields, :]*scale
        else:
            dset[0] = amp_mean*scale
            dset[1] = amp_stddev*scale

    elif extname.endswith("CHISQ"):
        
        amp_mean = fits_handler(input="chisq_c0001_k000001.fits", min=burnin, max=None, minchain=cmin, maxchain=cmax, chdir=chdir, output="map", fwhm=fwhm, nside=nside, zerospin=False, drop_missing=True, pixweight=None, command=np.mean, lowmem=False, write=False)
        #amp_stddev = fits_handler(input="chisq_c0001_k000001.fits", min=burnin, max=None, minchain=cmin, maxchain=cmax, chdir=chdir, output="map", fwhm=fwhm, nside=nside, zerospin=False, drop_missing=True, pixweight=None, command=np.std, lowmem=False, write=False)

        dset = np.zeros((len(types), hp.nside2npix(nside)))

        dset[0] = amp_mean[0, :]
        dset[1] = amp_mean[1, :]+amp_mean[2, :]
    else:
        print(f"Have not set up case for {extname}")

    #print(f"Shape of dset {dset.shape}")
    return dset

def get_header(extname, types, units, nside, polar, component, fwhm, nu_ref_t, nu_ref_p, procver, filename, bndctr, restfreq, bndwid,):
    stamp = f'Written {time.strftime("%c")}'

    header = []
    header.append(("DATE", stamp, "Time and date of creation.",))
    header.append(("PIXTYPE", "HEALPIX", "HEALPIX pixelisation.",))
    header.append(("COORDSYS", "GALACTIC"))
    header.append(("POLAR", polar))
    header.append(("BAD_DATA", hp.UNSEEN, "HEALPIX UNSEEN value.",))
    header.append(("METHOD", "COMMANDER", "COMMANDER sampling framework",))
    header.append(("AST-COMP", component))
    if extname == "FREQMAP":
        header.append(("FREQ", nu_ref_t))
    else:
        header.append(("FWHM", fwhm))
        header.append(("NU_REF_T", nu_ref_t))
        header.append(("NU_REF_P", nu_ref_p))

    header.append(("PROCVER", procver, "Release version"))
    header.append(("FILENAME", filename))
    if extname == "FREQMAP":
        # TODO are these correct?
        header.append(("BNDCTR", bndctr, "Formal Band Center",))
        header.append(("RESTFREQ", restfreq, "Effective Central Frequency",))
        header.append(("BNDWID", bndwid, "Effective Bandwidth",))
    return header



import healpy as hp
import matplotlib.pyplot as plt

from functools import partial
from .plottools import *


@u.quantity_input(freq=u.Hz, fwhm=(u.arcmin, u.rad, u.deg))
def gnom(
    input,
    lon=0,
    lat=0,
    comp=None,
    sig=0,
    freq=None,
    nside=None,
    size=20,
    sample=-1,
    min=None,
    max=None,
    ticks=None,
    rng=None,
    llabel=None,
    rlabel=None,
    unit=None,
    cmap=None,
    graticule=False,
    norm=None,
    cbar=True,
    cbar_pad=0.0,
    cbar_shrink=1,
    fwhm=0.0 * u.arcmin,
    remove_dip=False,
    remove_mono=False,
    fontsize=None,
    figsize=(3, 4),
    xsize=1000,
    darkmode=False,
    sub=None,
    hold=False,
    reuse_axes=False,
    gb=False,
    textbg_alpha=0.9
):
    """
    Gnomonic view plotting

    Parameters
    ----------
    input : ndarray, fits file path or cosmoglobe model object
        Map data input given as numpy array either 1d or index given by 'sig'.
        Also supports fits-file path string or cosmoglobe model.
        If cosmoglobe object is passed such as 'model', specify comp or freq.
    lon : float
        Longitude coordinate of the center of the figure in galactic coordinates
        default: 0
    lat : float
        Latitude coordinate of the center of the figure in galactic coordinates
        default: 0
    sig : str or int, optional
        Specify which signal to plot if ndim>1.
        default: None
    comp : string, optional
        Component label for automatic identification of plotting
        parameters based on information from autoparams.json
        default: None
    freq : astropy GHz, optional
        frequency in GHz needed for scaling maps when using a model object input
        default: None
    nside : int, optional
        optional ud_grading of the input data
        default: None
    size : float, optional
        Size of square in degrees
        default: 20
    sample : float, optional
        Specify sample if passing chain
        default: -1
    min : float, optional
        Min value
        default: None
    max : float, optional
        Max value
        default: None
    ticks : list or str, optional
        Ticks for colorbar
        default: None
    rng : float, optional
      Sets this value as min and max value. If specified, overwrites autodetector.
      default: None
    llabel : str, optional
        Sets the upper left title. Has LaTeX functionaliity (ex. $A_{s}$.)
        default: None
    rlabel : str, optional
        Sets the upper right title. Has LaTeX functionaliity (ex. $A_{s}$.)
        default: None
    unit : str, optional
        Unit label for colorbar
        default: None
    cmap : str, optional
        Colormap (ex. sunburst, planck, jet). Both matplotliib and cmasher
        available as of now. Also supports qualitative plotly map, [ex.
        q-Plotly-4 (q for qualitative 4 for max color)] Sets planck as default.
        default: None
    graticule : bool
        add graticule
        default: False
    norm : str, matplotlib norm object, optional
        if norm=='linear':
            normal
        if norm=='log':
            Uses log10 scale
            Normalizes data using a semi-logscale linear between -1 and 1.
            Autodetector uses this sometimes, you will be warned.
        SPECIFY linthresh for linear threshold of symlog!
        default: None
    cbar : bool, optional
        Toggles the colorbar
        cbar : True
    cbar_pad : float, optioinal
        default: 0.04
    cbar_shrink : float, optional
        default: 0.7
    fwhm : astropy arcmin/rad/deg, optional
        Optional map smoothing. FWHM of gaussian smoothing in arcmin.
        default: 0.0
    remove_dip : bool, optional
        If mdmask is specified, fits and removes a dipole.
        default: True
    remove_mono : bool, optional
        If mdmask is specified, fits and removes a monopole.
        default: True
    fontsize : int, optional
        Size of labels
        default: None
    figsize : touple, optional
        size of figure
        default: None
    xsize : int, optional
        width of figure in pixels
        default: 1000
    darkmode : bool, optional
        turn all axis elements white for optimal dark visualization
        default: False
    sub : int, scalar or sequence, optional
        Use only a zone of the current figure (same syntax as subplot).
        Default: None
    hold : bool, optional
        If True, replace the current Axes by a MollweideAxes.
        use this if you want to have multiple maps on the same
        figure.
        Default: False
    reuse_axes : bool, optional
        If True, reuse the current Axes (should be a MollweideAxes). This is
        useful if you want to overplot with a partially transparent colormap,
        such as for plotting a line integral convolution.
        Default: False

    mask : str path or np.ndarray, optional
        Apply a mask file to data
        default = None
    """
    default_fontsize= {
            "xlabel": 11,
            "ylabel": 11,
            "xtick_label": 8,
            "ytick_label": 8,
            "title": 12,
            "cbar_label": 11,
            "cbar_tick_label": 9,
            "llabel": 11,
            "rlabel": 11,
        }
    if fontsize is not None: default_fontsize.update(fontsize)

    # Set plotting rcParams
    set_style(darkmode)
    reso = size * 60 / xsize

    # Get data
    m, comp, freq, nside = get_data(input, sig, comp, freq, fwhm, nside=nside, sample=sample)
    m = hp.ma(m)
    if remove_dip:
        m = hp.remove_dipole(m, gal_cut=30, copy=True, verbose=True)
    if remove_mono:
        m = hp.remove_monopole(m, gal_cut=30, copy=True, verbose=True)

    proj = hp.projector.GnomonicProj(rot=[lon, lat, 0.0], coord="G", xsize=xsize, ysize=xsize, reso=reso)
    reproj_im = proj.projmap(m, vec2pix_func=partial(hp.vec2pix, nside))
    mask = (reproj_im == hp.UNSEEN)
    reproj_im = np.ma.masked_array(reproj_im, mask=mask)

    # Pass all your arguments in, return parsed plotting parameters
    params = get_params(
        data=reproj_im,
        comp=comp,
        sig=sig,
        rlabel=rlabel,
        llabel=llabel,
        unit=unit,
        ticks=ticks,
        min=min,
        max=max,
        rng=rng,
        norm=norm,
        cmap=cmap,
        freq_ref=freq,
        width=figsize[0],
        nside=nside,
    )

    # Semi-log normalization
    if params["norm"] == "log":
        params["data"], params["ticks"] = apply_logscale(params["data"], params["ticks"], linthresh=1)
    
    cmap = load_cmap(params["cmap"])
    cmap.set_bad('gray')
    
    fig, ax = make_fig(figsize, None, hold, sub, reuse_axes)
    image = plt.imshow(
        params["data"],
        origin="lower",
        interpolation="nearest",
        vmin=np.min(params["ticks"]),
        vmax=np.max(params["ticks"]),
        cmap=cmap,
    )

    # Galaxy-brain inverse of data color
    if gb:
        colors = image.cmap(image.norm((params["data"])))
        llabel_color = (1, 1, 1, 2) - np.mean(colors[int(xsize * 0.9) : int(xsize * 0.95), int(xsize * 0.05) : int(xsize * 0.15)], axis=(0, 1))
        rlabel_color = (1, 1, 1, 2) - np.mean(colors[int(xsize * 0.9) : int(xsize * 0.95), int(xsize * 0.85) : int(xsize * 0.95)], axis=(0, 1))
        bbox = dict()
    else:
        llabel_color = 'k'
        rlabel_color = 'k'
        backgroundcolor = 'white'
        bbox = dict(facecolor='white', alpha=textbg_alpha)

    plt.xticks([])
    plt.yticks([])
    plt.text(
        0.95,
        0.95,
        params["rlabel"],
        color=rlabel_color,
        va="top",
        ha="right",
        transform=ax.transAxes,
        fontsize=default_fontsize["rlabel"],
        bbox=bbox,
    )
    plt.text(
        0.05,
        0.95,
        params["llabel"],
        color=llabel_color,
        va="top",
        transform=ax.transAxes,
        fontsize=default_fontsize["llabel"],
        bbox=bbox,
    )
    if cbar:
        apply_colorbar(
            plt.gcf(),
            plt.gca(),
            image,
            params["ticks"],
            params["ticklabels"],
            params["unit"],
            fontsize=default_fontsize,
            linthresh=1,
            norm=params["norm"],
            cbar_pad=cbar_pad,
            cbar_shrink=cbar_shrink,
        )

    if graticule:
        hp.graticule()
    return image, params



from matplotlib import rcParams
import matplotlib.patheffects as path_effects
import matplotlib.ticker as ticker
import matplotlib.pyplot as plt
from cycler import cycler
from tqdm import tqdm
import numpy as np
import healpy as hp

import cosmoglobe.release.tools as tls
def Spectrum(pol, long, darkmode, png, foregrounds, masks, nside, cmap="Plotly"):
    params = {'savefig.dpi'        : 300, # save figures to 300 dpi
              'xtick.top'          : False,
              'ytick.right'        : True, #Set to false
              'axes.spines.top'    : True, #Set to false
              'axes.spines.bottom' : True,
              'axes.spines.left'   : True,
              'axes.spines.right'  : True, #Set to false@
              'axes.grid.axis'     : 'y',
              'axes.grid'          : False,
              'ytick.major.size'   : 10,
              'ytick.minor.size'   : 5,
              'xtick.major.size'   : 10,
              'xtick.minor.size'   : 5,
              'xtick.major.pad'    : 8, 
              'ytick.major.width'   : 1.5,
              'ytick.minor.width'   : 1.5,
              'xtick.major.width'   : 1.5,
              'xtick.minor.width'   : 1.5,
              'axes.linewidth'      : 1.5,
              'mathtext.fontset': 'dejavusans',
              #'ytick.major.size'   : 6,
              #'ytick.minor.size'   : 3,
              #'xtick.major.size'   : 6,
              #'xtick.minor.size'   : 3,
              
    }
    black = 'k'
    if darkmode:
        brokecol="white"
        grey="#C0C0C0"
    else:
        brokecol=black
        grey="grey"
    colors=['#636EFA', '#EF553B', '#00CC96', '#AB63FA', '#FFA15A', '#19D3F3', '#FF6692', '#B6E880', '#FF97FF', '#FECB52', grey,]
    rcParams['axes.prop_cycle'] = cycler(color=colors)
    blue, red, green, purple, orange, teal, lightred, lightgreen, pink, yellow, grey = ("C0","C1","C2","C3","C4","C5","C6","C7","C8","C9","C10")

    if darkmode:
        rcParams['text.color']   = 'white'   # axes background color
        rcParams['axes.facecolor']   = 'white'   # axes background color
        rcParams['axes.edgecolor' ]  = 'white'   # axes edge color
        rcParams['axes.labelcolor']   = 'white'
        rcParams['xtick.color']    = 'white'      # color of the tick labels
        rcParams['ytick.color']    = 'white'      # color of the tick labels
        rcParams['grid.color'] =   'white'   # grid color
        rcParams['legend.facecolor']    = 'inherit'   # legend background color (when 'inherit' uses axes.facecolor)
        rcParams['legend.edgecolor']= 'white' # legend edge color (when 'inherit' uses axes.edgecolor)
        black = 'white'
    rcParams.update(params)
    
    # ---- Figure parameters ----
    if long:
        xmin, xmax = (0.25, 4000)
    if pol:
        ymin, ymax = (1.001e-3, 2e2)
        if long:
            #xmin, xmax = (1, 3000)
            ymax15, ymax2 = (ymax+100, 1e7)
        else:
            xmin, xmax = (9, 1500)
    else:
        ymin, ymax = (0.05, 7e2)
        if long:
            #xmin, xmax = (0.3, 4000)
            ymax15, ymax2 = (ymax+500, 1e7)
        else:
            xmin, xmax = (9, 1500)

    if long:    

        # Figure
        ratio = 5
        w, h = (16,8)
        fig, (ax2, ax) = plt.subplots(2,1,sharex=True,figsize=(w,h),gridspec_kw = {'height_ratios':[1, ratio]})

        ax2.spines['bottom'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax2.tick_params(labelbottom=False)
        ax2.xaxis.set_ticks_position('none')

        # ---- Adding broken axis lines ----
        d = .005  # how big to make the diagonal lines in axes coordinates
        kwargs = dict(transform=ax2.transAxes, color=brokecol, clip_on=False)
        ax2.plot((-d, +d), (-d*ratio, + d*ratio), **kwargs)        # top-left diagonal
        ax2.plot((1 - d, 1 + d), (-d*ratio, +d*ratio), **kwargs)  # top-right diagonal
        kwargs.update(transform=ax.transAxes)  # switch to the bottom axes
        ax.plot((-d, +d), (1 - d, 1 + d), **kwargs)  # bottom-left diagonal
        ax.plot((1 - d, 1 + d), (1 - d, 1 + d), **kwargs)  # bottom-right diagonal
    
        # textsize
        freqtext = 16
        fgtext = 18

    else:
        ymax2=ymax
        ymax15=ymax
        w, h = (12,8)
        fig, ax = plt.subplots(1,1,figsize=(w,h))
        
        freqtext = 20
        fgtext = 20

    # Spectrum parameters
    field = 1 if pol else 0
    nu  = np.logspace(np.log10(0.1),np.log10(5000),1000)
    npix = hp.nside2npix(nside)
    # Read masks
    m = np.ones((len(masks), npix))
    for i, mask in enumerate(masks):
        # Read and ud_grade mask
        if mask:
            m_temp = hp.read_map(mask, field=0, dtype=None, verbose=False)
            if hp.npix2nside(len(m_temp)) != nside:
                m[i] = hp.ud_grade(m_temp, nside)
                m[i,m[i,:]>0.5] = 1 # Set all mask values to integer    
                m[i,m[i,:]<0.5] = 0 # Set all mask values to integer   
            else:
                m[i] = m_temp


    # Get indices of smallest mask
    idx = m[np.argmax(np.sum(m, axis=1)), :] > 0.5
    skyfracs = np.sum(m,axis=1)/npix*100
    print(f"Using sky fractions {skyfracs}%")
    # Looping over foregrounds and calculating spectra
    print("Foreground items:", foregrounds.items())
    i = 0
    add_error = True
    for fg in foregrounds.keys():
        if not fg == "Sum fg.":
            if fg.startswith("CO"): # get closest thing to ref freq
                foregrounds[fg]["params"][-2], _ = find_nearest(nu, foregrounds[fg]["params"][-2])

            foregrounds[fg]["spectrum"] = getspec(nu*1e9, fg, foregrounds[fg]["params"], foregrounds[fg]["function"], field, nside, npix, idx, m,)
            foregrounds[fg]["spectrum_mean"]= np.mean(foregrounds[fg]["spectrum"],axis=0)
            if add_error and foregrounds[fg]["spectrum"].shape[0]>1 and not fg.startswith("CO"):
                thresh=0.1                    
                alpha=0.5
                foregrounds[fg]["spectrum"][0] = foregrounds[fg]["spectrum"][0]*(1-np.exp(-(abs(foregrounds[fg]["spectrum"][0]/thresh)**alpha)))
                foregrounds[fg]["spectrum"][1] = foregrounds[fg]["spectrum"][1]/(1-np.exp(-(abs(foregrounds[fg]["spectrum"][1]/thresh)**alpha)))

        if foregrounds[fg]["sum"]:
            if i==0:
                if foregrounds[fg]["spectrum"].shape[0] == 1:
                    # special case where first summed is 1d
                    foregrounds["Sum fg."]["spectrum"] = np.concatenate((foregrounds[fg]["spectrum"],foregrounds[fg]["spectrum"])).copy()
                else:
                    foregrounds["Sum fg."]["spectrum"] = foregrounds[fg]["spectrum"].copy()
            else:
                foregrounds["Sum fg."]["spectrum"] += foregrounds[fg]["spectrum"]
            i+=1

    # ---- Plotting foregrounds and labels ----
    j=0
    for label, fg in foregrounds.items(): # Plot all fgs except sumf
        print(f"plotting {label}")
        if fg["gradient"]:
            k = fg["spectrum"].shape[0] - 1
            gradient_fill_between(ax, nu, fg["spectrum"][k]*1e-2, fg["spectrum"][k], color=fg["color"])
        else:
            if label == "Sum fg.":
                ax.loglog(nu,fg["spectrum"][0], linestyle=fg["linestyle"], linewidth=2, color=fg["color"])
                if long:
                    ax2.loglog(nu,fg["spectrum"][0], linestyle=fg["linestyle"], linewidth=2, color=fg["color"])
                k = 0
                try:
                    ax.loglog(nu,fg["spectrum"][1], linestyle=fg["linestyle"], linewidth=2, color=fg["color"])
                    if long:
                        ax2.loglog(nu,fg["spectrum"][1], linestyle=fg["linestyle"], linewidth=2, color=fg["color"])
                    k=1
                except:
                    pass
            elif label.startswith("CO"):
                lfreq = nu[np.argmax(fg["spectrum"][0])]
                if fg["spectrum"].shape[0] > 1:
                    ax.loglog([lfreq,lfreq],[max(fg["spectrum"][0]), max(fg["spectrum"][1])], linestyle=fg["linestyle"], linewidth=4, color=fg["color"],zorder=1000)
                    k=1
                else:
                    k=0
                    ax.bar(lfreq, fg["spectrum"][0], color=black,)
            else:
                if fg["spectrum"].shape[0] == 1:
                    ax.loglog(nu,fg["spectrum"][0], linestyle=fg["linestyle"], linewidth=4, color=fg["color"])
                    if long:
                        ax2.loglog(nu,fg["spectrum"][0], linestyle=fg["linestyle"], linewidth=4, color=fg["color"])
                    k = 0
                else:
                    #gradient_fill(nu, fg["spectrum"][0], fill_color=fg["color"], ax=ax, alpha=0.5, linewidth=0.0,)
                    
                    #ax.loglog(nu,fg["spectrum_mean"], linestyle=fg["linestyle"], linewidth=4, color=fg["color"])
                    ax.fill_between(nu,fg["spectrum"][0],fg["spectrum"][1], color=fg["color"],alpha=0.8)

                    if long:
                        #ax2.loglog(nu,fg["spectrum_mean"], linestyle=fg["linestyle"], linewidth=4, color=fg["color"])
                        ax2.fill_between(nu,fg["spectrum"][0],fg["spectrum"][1], color=fg["color"], alpha=0.8)
                    k = 1

        if label == "Thermal Dust" and fg["spectrum"].shape[0]>1:
            _, fsky_idx = find_nearest(nu, 900)
            ax.annotate(r"$f_{sky}=$"+"{:d}%".format(int(skyfracs[1])), xy=(nu[fsky_idx], fg["spectrum"][1][fsky_idx]), ha="center", va="bottom", fontsize=fgtext, color="grey", xytext=(0,5), textcoords="offset pixels",path_effects=[path_effects.withSimplePatchShadow(alpha=0.8,offset=(0.5, -0.5)),])
            ax.annotate(r"$f_{sky}=$"+"{:d}%".format(int(skyfracs[0])), xy=(nu[fsky_idx], fg["spectrum"][0][fsky_idx]), ha="center", va="top", fontsize=fgtext, color="grey", xytext=(0,-15), textcoords="offset pixels",path_effects=[path_effects.withSimplePatchShadow(alpha=0.8,offset=(0.5, -0.5)),])
       
        if label.startswith("CO"):
            ax.text(lfreq, np.max(fg["spectrum"][k])*0.5, label, color=fg["color"], alpha=0.7, ha='right',va='center',rotation=90,fontsize=fgtext, path_effects=[path_effects.withSimplePatchShadow(alpha=0.8, offset=(1, -1))], zorder=1000)
        else:
            print(label, fg["position"])
            x0, idx1 = find_nearest(nu, fg["position"])
            idx2 = idx1+2
            x1 = nu[idx2] 
            # idx2 = find_nearest(nu, fg["position"]**1.05)
            y0 = fg["spectrum"][k][idx1]
            y1 = fg["spectrum"][k][idx2]
            if y0>ymax:
                aspect_ratio = w/(h/6) if long else w/h
                datascaling  = np.log(xmin/xmax)/np.log(ymax15/ymax2)
            else:
                aspect_ratio = w/(h*5/6) if long else w/h
                datascaling  = np.log(xmin/xmax)/np.log(ymin/ymax)
            rotator = (datascaling/aspect_ratio)
            alpha = np.arctan(np.log(y1/y0)/np.log(x1/x0)*rotator)
            rotation =  np.rad2deg(alpha)#*rotator
            if y0>ymax and long:
                ax2.annotate(label, xy=(x0,y0), xytext=(0,5), textcoords="offset pixels",  rotation=rotation, rotation_mode='anchor', fontsize=fgtext, color=fg["color"], path_effects=[path_effects.withSimplePatchShadow(alpha=0.8,offset=(1, -1)),], horizontalalignment="center")
            else:
                ax.annotate(label, xy=(x0,y0), xytext=(0,5), textcoords="offset pixels",  rotation=rotation, rotation_mode='anchor', fontsize=fgtext, color=fg["color"], path_effects=[path_effects.withSimplePatchShadow(alpha=0.8,offset=(1, -1)),], horizontalalignment="center")
    
    # ---- Data band ranges ----
    if long:
        yscaletext = 0.70
        yscaletextup = 1.2
    else:
        yscaletextup = 1.03
        yscaletext = 0.90

    # TODO add these as args?
    haslam = True
    chipass = True
    spass = True
    cbass = True
    quijote = True
    wmap = True
    planck = True
    dirbe = True
    bicep=False
    spider=False
    SPT=False
    act=False
    so = False
    dirbepos=1000 if long else 950
    spass_special = ymax2*yscaletext if pol else ymin*yscaletextup 
    databands = {"Haslam":  {"0.4\nHaslam": {"pol": False, "show": haslam, "position": [.408, ymin*yscaletextup],  "range": [.406,.410], "color": green,}},
                 "S-PASS":  {"2.3\nS-PASS":  {"pol": True, "show": spass,  "position": [2.35, spass_special],  "range": [2.1,2.4], "color": green,}},
                 "C-BASS":  {"5.0\nC-BASS":   {"pol": True, "show": cbass,  "position": [5., ymax2*yscaletext],    "range": [4.,6.], "color": blue,}},
                 "CHI-PASS":{"CHI-PASS\n1.4":{"pol": False, "show": chipass,"position": [1.35, ymin*yscaletextup],"range": [1.3945-0.064/2, 1.3945+0.064/2], "color": lightred,}},
                 "QUIJOTE": {"11":    {"pol": True, "show": quijote,"position": [10.5, ymax2*yscaletext],    "range":  [10.,12.], "color": red,},
                             "13":             {"pol": True, "show": quijote, "position": [13.5, ymax2*yscaletext], "range":  [12.,14.], "color": red,},
                             "17\nQUIJOTE":             {"pol": True, "show": quijote, "position": [17, ymax2*yscaletext], "range":  [16.,18.], "color": red,},
                             "19":             {"pol": True, "show": quijote, "position": [21, ymax2*yscaletext], "range":  [18.,21.], "color": red,},
                             ".\n31":             {"pol": True, "show": quijote, "position": [31, ymax2*yscaletext], "range":  [26.,36.], "color": red,},
                             ".\n41":             {"pol": True, "show": quijote, "position": [42, ymax2*yscaletext], "range":  [35.,47.], "color": red,}},
                 "Planck":  {"30":          {"pol": True, "show": planck, "position": [27,  ymax2*yscaletext], "range": [23.9,34.5],"color": orange,},      # Planck 30
                             "44":          {"pol": True, "show": planck, "position": [40,  ymax2*yscaletext], "range": [39,50]    ,"color": orange,},      # Planck 44
                             "70":          {"pol": True, "show": planck, "position": [60,  ymax2*yscaletext], "range": [60,78]    ,"color": orange,},      # Planck 70
                             "100": {"pol": True, "show": planck, "position": [90,  ymax2*yscaletext], "range": [82,120]   ,"color": orange,},      # Planck 100
                             "143\nPlanck":         {"pol": True, "show": planck, "position": [130, ymax2*yscaletext], "range": [125,170]  ,"color": orange,},      # Planck 143
                             "217":         {"pol": True, "show": planck, "position": [195, ymax2*yscaletext], "range": [180,265]  ,"color": orange,},      # Planck 217
                             "353":         {"pol": True, "show": planck, "position": [320, ymax2*yscaletext], "range": [300,430]  ,"color": orange,},      # Planck 353
                             "545":         {"pol": False, "show": planck, "position": [490, ymax2*yscaletext], "range": [450,650]  ,"color": orange,},      # Planck 545
                             "857":         {"pol": False, "show": planck, "position": [730, ymax2*yscaletext], "range": [700,1020] ,"color": orange,}},      # Planck 857
                 "DIRBE":   {"DIRBE\n1250":  {"pol": False, "show": dirbe, "position": [dirbepos, ymin*yscaletextup], "range": [1000,1540] , "color": red,},     # DIRBE 1250
                             "2140":         {"pol": False, "show": dirbe, "position": [1650, ymin*yscaletextup], "range": [1780,2500] , "color": red,},     # DIRBE 2140
                             "3000":         {"pol": False, "show": dirbe, "position": [2500, ymin*yscaletextup], "range": [2600,3500] , "color": red,}},     # DIRBE 3000
                 "WMAP":    {"K": {"pol": True, "show": wmap, "position": [21.8, ymin*yscaletextup], "range": [21,25.5], "color": teal,}, 
                             "WMAP\nKa":      {"pol": True, "show": wmap, "position": [31.5, ymin*yscaletextup], "range": [30,37], "color": teal,},
                             "Q":       {"pol": True, "show": wmap, "position": [39.,  ymin*yscaletextup], "range": [38,45], "color": teal,}, 
                             "V":       {"pol": True, "show": wmap, "position": [58.,  ymin*yscaletextup], "range": [54,68], "color": teal,}, 
                             "W":       {"pol": True, "show": wmap, "position": [90.,  ymin*yscaletextup], "range": [84,106], "color": teal,}}, 
                 "BICEPx/KECK":    {"95\nBICEPx/KECK": {"pol": True, "show": bicep, "position": [21.8, ymin*yscaletextup], "range": [21,25.5], "color": teal,}, 
                                    "150":      {"pol": True, "show": bicep, "position": [31.5, ymin*yscaletextup], "range": [30,37], "color": teal,},
                                    "220":       {"pol": True, "show": bicep, "position": [39.,  ymin*yscaletextup], "range": [38,45], "color": teal,}},
                 "SPT":    {"90": {"pol": True, "show": SPT, "position": [21.8, ymin*yscaletextup], "range": [21,25.5], "color": teal,}, 
                            "150":      {"pol": True, "show": SPT, "position": [31.5, ymin*yscaletextup], "range": [30,37], "color": teal,},
                            "220\nSPT":       {"pol": True, "show": SPT, "position": [39.,  ymin*yscaletextup], "range": [38,45], "color": teal,}}, 
                 "SPIDER":    {"100": {"pol": True, "show": spider, "position": [21.8, ymin*yscaletextup], "range": [21,25.5], "color": teal,}, 
                               "150":      {"pol": True, "show": spider, "position": [31.5, ymin*yscaletextup], "range": [30,37], "color": teal,},
                               "280\nSPIDER":       {"pol": True, "show": spider, "position": [39.,  ymin*yscaletextup], "range": [38,45], "color": teal,}}, 
                 "SO":     {"27":           {"pol": True, "show": so, "position": [27,  ymax2*yscaletext], "range": [23.9,34.5],"color": orange,},      # Planck 30
                            "39\nSO":       {"pol": True, "show": so, "position": [40,  ymax2*yscaletext], "range": [39,50]    ,"color": orange,},      # Planck 44
                             "93":          {"pol": True, "show": so, "position": [60,  ymax2*yscaletext], "range": [60,78]    ,"color": orange,},      # Planck 70
                             "145":         {"pol": True, "show": so, "position": [90,  ymax2*yscaletext], "range": [82,120]   ,"color": orange,},      # Planck 100
                             "225":         {"pol": True, "show": so, "position": [130, ymax2*yscaletext], "range": [125,170]  ,"color": orange,},      # Planck 143
                             "280":         {"pol": True, "show": so, "position": [195, ymax2*yscaletext], "range": [180,265]  ,"color": orange,}},      # Planck 217
                 "ACT":     {"28":          {"pol": True, "show": act, "position": [27,  ymax2*yscaletext], "range": [23.9,34.5],"color": orange,},      # Planck 30
                             "41":          {"pol": True, "show": act, "position": [40,  ymax2*yscaletext], "range": [39,50]    ,"color": orange,},      # Planck 44
                             "90\nACT":          {"pol": True, "show": act, "position": [60,  ymax2*yscaletext], "range": [60,78]    ,"color": orange,},      # Planck 70
                             "150": {"pol": True, "show": act, "position": [90,  ymax2*yscaletext], "range": [82,120]   ,"color": orange,},      # Planck 100
                             "220":         {"pol": True, "show": act, "position": [130, ymax2*yscaletext], "range": [125,170]  ,"color": orange,}},      # Planck 143            
    }
    
    bands2 = {"Planck": [30,44,70,100,143,216,353,545,857],
              "WMAP": [23,33,41,61,94],
              "S-PASS": [2.303],
              "C-BASS": [5.0],
              "CHI-PASS": [1.4],
              "QUIJOTE": [11, 13, 17, 19, 31, 41],
              "LiteBIRD": [40, 50, 60, 68, 78, 89, 100, 119, 140, 166, 195, 235, 280, 337, 402],
              "GroundBIRD": [145,220],
              "LSPE": [43,240],
              "QUBIC": [150, 220],
              "SO": [27,39,93,145,225,280],
              "ACT": [28,41,90, 150, 220,],
              "PICO": [21, 25, 30, 36, 43, 52, 62, 75, 90, 108, 129, 155, 186, 223, 268, 321, 385, 462, 555, 666, 799],
              "SPT": [90,150,220],
              "SPIDER": [100,150,280],
              "BICEP": [95,150,220],
              "CMB-S4": [30,40,85,95,145,155,220,270,20,27,39,93,145,225,278,27,39,93,145,225,278],
              "CMB-BHARAT": [60,600],
          }

    # Set databands from dictonary
    yrows=np.logspace(0,2,8)[::-1]
    cm = plt.cm.get_cmap('tab20')
    if False:
        c=0
        k=0
        for label, freqs in bands2.items():
            va="top"
            xpos=max(freqs)
            if xpos<10:
                ax2.text(xpos,ymax2*yscaletext, label, color=cm.colors[c], va=va, ha="center", size=freqtext, path_effects=[path_effects.withSimplePatchShadow(alpha=0.8, offset=(1,-1))])
            else:
                if k>7:               
                    ax.text(4,yrows[k-8], label, color=cm.colors[c], va=va, ha="left", size=freqtext, path_effects=[path_effects.withSimplePatchShadow(alpha=0.8, offset=(1,-1))])
                else:
                    ax.text(1.4,yrows[k], label, color=cm.colors[c], va=va, ha="left", size=freqtext, path_effects=[path_effects.withSimplePatchShadow(alpha=0.8, offset=(1,-1))])
                k+=1

            for nu in freqs:
                ax.axvline(nu,color=cm.colors[c], alpha=0.4, zorder=0)
                if long:
                    ax2.axvline(nu,color=cm.colors[c],alpha=0.4,zorder=0)
            c+=1
    else:
        for experiment, bands in databands.items():
            for label, band in bands.items():
                if band["show"]:
                    #if label == "353" and not pol: continue # SPECIFIC FOR BP SETUP
                    if pol and not band["pol"]:
                        continue # Skip non-polarization bands
                    if band["position"][0]>=xmax or band["position"][0]<=xmin:
                        continue # Skip databands outside range

                    bottombands = ["WMAP", "CHI-PASS", "DIRBE", "Haslam"]
                    if not pol: 
                        bottombands.append("S-PASS")
                    va = "bottom" if experiment in bottombands else "top" # VA for WMAP on bottom
                    ha = "left" if experiment in ["Planck", "WMAP", "DIRBE",] else "center"
                    ax.axvspan(*band["range"], color=band["color"], alpha=0.3, zorder=-20, label=experiment)
                    if long:
                        ax2.axvspan(*band["range"], color=band["color"], alpha=0.3, zorder=-20, label=experiment)
                        if experiment in bottombands:
                            ax.text(*band["position"], label, color=band["color"], va=va, ha=ha, size=freqtext, path_effects=[path_effects.withSimplePatchShadow(alpha=0.8, offset=(1,-1))])
                        else:
                            ax2.text(*band["position"], label, color=band["color"], va=va, ha=ha, size=freqtext, path_effects=[path_effects.withSimplePatchShadow(alpha=0.8, offset=(1,-1))])
                    else:
                        ax.text(*band["position"], label, color=band["color"], va=va, ha=ha, size=freqtext, path_effects=[path_effects.withSimplePatchShadow(alpha=0.8, offset=(1,-1))])

    # ---- Axis stuff ----
    lsize=20
    """
    majorticks = minorticks = []
    majorticks_ = [1,10,100,1000]
    minorticks_ = [0.3,3,30,300,3000]
    for i, tick in enumerate(minorticks_):
        if tick>=xmin and tick<=xmax:
            minorticks.append(tick)
    for i, tick in enumerate(majorticks_):
        if tick>=xmin and tick<=xmax:
            majorticks.append(tick)
                        

    """

    ticks = []
    ticks_ = [0.3,1,3,10,30,100,300,1000,3000]

    for i, tick in enumerate(ticks_):
        if tick>=xmin and tick<=xmax:
            ticks.append(tick)
    ax.set(xscale='log', yscale='log', ylim=(ymin, ymax), xlim=(xmin,xmax),xticks=ticks, xticklabels=ticks)
    ax.xaxis.set_major_formatter(ticker.ScalarFormatter(useMathText=True))
    ax.tick_params(axis='both', which='major', labelsize=lsize, direction='in')
    ax.tick_params(which="both",direction="in")
    ax.tick_params(axis='y', labelrotation=90) 
    ax.set_yticklabels([fmt(x,1) for x in ax.get_yticks()], va="center")
    if long:
        ax2.set(xscale='log', yscale='log', ylim=(ymax15, ymax2), xlim=(xmin,xmax), yticks=[1e4,1e6,], xticks=ticks, xticklabels=ticks)
        ax2.tick_params(axis='both', which='major', labelsize=lsize, direction='in')
        ax2.tick_params(which="both",direction="in")
        ax2.tick_params(axis='y', labelrotation=90,) 
        ax2.set_yticklabels([fmt(x,1) for x in ax2.get_yticks()], va="center")

    # Axis labels
    sax = fig.add_subplot(111, frameon=False)
    plt.tick_params(
        labelcolor="none",
        top=False,
        bottom=False,
        left=True,
        right=False,
        width=0.0,
    )
    if pol:
        sax.set_ylabel(r"RMS polarization amplitude [$\mu\mathrm{K}_{\mathrm{RJ}}$]",fontsize=lsize)
    else:
        sax.set_ylabel(r"RMS brightness temperature [$\mu\mathrm{K}_{\mathrm{RJ}}$]",fontsize=lsize)
    sax.set_xlabel(r"Frequency [GHz]",fontsize=lsize)
    plt.subplots_adjust(wspace=0.0, hspace=0.02)

    #ax.legend(loc=6,prop={'size': 20}, frameon=False)

    # ---- Plotting ----
    #plt.tight_layout(h_pad=0.3)
    filename = "spectrum"
    filename += "_pol" if pol else ""
    filename += "_long" if long else ""
    filename += "_darkmode" if darkmode else ""
    filename += ".png" if png else ".pdf"
    print("Plotting {}".format(filename))
    plt.savefig(filename, bbox_inches='tight',  pad_inches=0.02, transparent=True)

def fmt(x, pos):
    """
    Format color bar labels
    """
    a, b = f"{x:.2e}".split("e")
    b = int(b)
    if float(a) == 1.00:
        return r"$10^{"+str(b)+"}$"
    elif float(a) == -1.00:
        return r"$-10^{"+str(b)+"}$"
    else:
        return r"$"+str(a)+"\cdot 10^{"+str(b)+"}$"

	
def gradient_fill(x, y, fill_color=None, ax=None, alpha=1.0, invert=False, **kwargs):
    """
    Plot a line with a linear alpha gradient filled beneath it.

    Parameters
    ----------
    x, y : array-like
        The data values of the line.
    fill_color : a matplotlib color specifier (string, tuple) or None
        The color for the fill. If None, the color of the line will be used.
    ax : a matplotlib Axes instance
        The axes to plot on. If None, the current pyplot axes will be used.
    Additional arguments are passed on to matplotlib's ``plot`` function.

    Returns
    -------
    line : a Line2D instance
        The line plotted.
    im : an AxesImage instance
        The transparent gradient clipped to just the area beneath the curve.
    """
    import matplotlib.colors as mcolors
    from matplotlib.patches import Polygon

    if ax is None:
        ax = plt.gca()

    line, = ax.plot(x, y, **kwargs)
    if fill_color is None:
        fill_color = line.get_color()

    zorder = line.get_zorder()
    #alpha = line.get_alpha()
    alpha = 1.0 if alpha is None else alpha

    z = np.empty((100, 1, 4), dtype=float)
    rgb = mcolors.colorConverter.to_rgb(fill_color)
    z[:,:,:3] = rgb
    z[:,:,-1] = np.linspace(0, alpha, 100)[:,None]

    xmin, xmax, ymin, ymax = x.min(), x.max(), y.min(), y.max()
    if invert:
        ymin,ymax = (ymax,ymin)
    im = ax.imshow(z, aspect='auto', extent=[xmin, xmax, ymin, ymax],
                   origin='lower', zorder=zorder)

    xy = np.column_stack([x, y])
    xy = np.vstack([[xmin, ymin], xy, [xmax, ymin], [xmin, ymin]])
    clip_path = Polygon(xy, facecolor='none', edgecolor='none', closed=True)
    ax.add_patch(clip_path)
    im.set_clip_path(clip_path)

    ax.autoscale(True)
    return line, im


def gradient_fill_between(ax, x, y1, y2, color="#ffa15a", alpha=0.8):
    N = 100
    y = np.zeros((N, len(y1)))
    for i in range(len(y1)):
        y[:,i] = np.linspace(y1[i], y2[i], N)

    alpha=np.linspace(0,alpha,100)**1.5
    for i in range(1,N):
        ax.fill_between(x, y[i-1], y[i], color=color, alpha=alpha[i], zorder=-10)
        ax.set_rasterization_zorder(-1)


def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx], idx


# This function calculates the intensity spectra
# Alternative 1 uses 2 masks to calculate spatial variations
# Alternative 2 uses only scalar values
def getspec(nu, fg, params, function, field, nside, npix, idx, m):
    val = []
    #print(fg)
    # Alternative 1
    if any([str(x).endswith(".fits") for x in params]) or any([isinstance(x,np.ndarray) for x in params]):
        if fg == "Spinning Dust":
            from pathlib import Path
            ame_template = Path(__file__).parent / "spdust2_cnm.dat"
            fnu, f_ = np.loadtxt(ame_template, unpack=True)
            fnu *= 1e9
            field = 0

        temp = []
        nsides = []
        # Read all maps and record nsides
        
        for i, p in enumerate(params):
            if str(p).endswith(".fits"):
                if field==1 and i==0: # If polarization amplitude map
                    s1 = hp.read_map(p, field=1, dtype=None, verbose=False)
                    s2 = hp.read_map(p, field=2, dtype=None, verbose=False)
                    p = np.sqrt(s1**2+s2**2)
                else:
                    p = hp.read_map(p, field=field, dtype=None, verbose=False)
                nsides.append(hp.npix2nside(len(p)))
            elif isinstance(p, np.ndarray):
                if not fg == "Spinning Dust":
                    if field==1 and i==0:
                        p = np.sqrt(p[1]**2+p[2]**2)
                    elif p.ndim > 1 and p.shape[0]>1:
                        p = p[field]
                nsides.append(hp.npix2nside(len(p)))
            else:
                nsides.append(0)
            temp.append(p)  


        # Create dataset and convert to same resolution
        params = np.zeros(( len(params), npix ))
        for i, t in enumerate(temp):
            if nsides[i] == 0:
                params[i,:] = t
            elif nsides[i] != nside:
                params[i,:] = hp.ud_grade(t, nside)
            else:
                params[i,:] = t
        # Only calculate outside masked region    
        N = 1000
        map_ = np.zeros((N, npix))

        for i, nu_ in enumerate(tqdm(nu, desc = fg, ncols=80)):
            if fg == "Spinning Dust":
                map_[i, idx] = getattr(tls, function)(nu_, *params[:,idx], fnu, f_) #fgs.fg(nu, *params[pix])
            else:
                map_[i, idx] = getattr(tls, function)(nu_, *params[:,idx]) #fgs.fg(nu, *params[pix])

        # Apply mask to all frequency points
        # calculate mean 
        rmss = []
        for i in range(2):
            n = np.sum(m[i])            
            masked = hp.ma(map_)
            masked.mask = np.logical_not(m[i])
            mono = masked.mean(axis=1)
            masked -= mono.reshape(-1,1)
            rms = np.sqrt( ( (masked**2).sum(axis=1) ) /n)
            val.append(rms)

        vals = np.sort(np.array(val), axis=0) 
    else:
        # Alternative 2
        val = getattr(tls, function)(nu, *params) #fgs.fg(nu, *params))
        #vals = np.stack((val, val),)
        vals = val.reshape(1,-1)
    return vals



from __future__ import annotations

from typing import Dict, Iterator, List, Literal, Optional, Union

import healpy as hp
import numpy as np
from astropy.units import Quantity
from astropy.units.core import UnitsError

from cosmoglobe.h5.chain import Chain
from cosmoglobe.sky._base_components import (
    DiffuseComponent,
    LineComponent,
    PointSourceComponent,
    SkyComponent,
)
from cosmoglobe.sky._component_factory import get_components_from_chain
from cosmoglobe.sky._constants import (
    DEFAULT_BEAM_FWHM,
    DEFAULT_GAL_CUT,
    DEFAULT_OUTPUT_UNIT_STR,
)
from cosmoglobe.sky._exceptions import (
    ComponentError,
    ComponentNotFoundError,
    NsideError,
)
from cosmoglobe.sky._units import Unit
from cosmoglobe.sky.components._labels import SkyComponentLabel
from cosmoglobe.sky.cosmoglobe import cosmoglobe_registry


class SkyModel:
    r"""Sky model representing an initialized instance of the Cosmoglobe Sky Model.

    This class acts as a container for the various components making up
    the Cosmoglobe Sky Model and provides methods to simulate the sky.
    The primary use case of this class is to call its ``__call__``
    method, which simulates the sky at a single frequency :math:`\nu`,
    or integrated over a bandpass :math:`\tau`.

    Examples
    --------
    >>> import cosmoglobe
    >>> chain = cosmoglobe.get_test_chain()
    >>> model = cosmoglobe.SkyModel.from_chain(chain, nside=256)
    >>> print(model)
    SkyModel(
        version: BeyondPlanck
        nside: 256
        components(
            (ame): SpinningDust(freq_peak)
            (cmb): CMB()
            (dust): ModifiedBlackbody(beta, T)
            (ff): LinearOpticallyThin(T_e)
            (radio): AGNPowerLaw(alpha)
            (synch): PowerLaw(beta)
        )
    )

    Simulating the full sky emission at some frequency, given a beam
    FWHM:

    >>> import astropy.units as u
    >>> model(50*u.GHz, fwhm=30*u.arcmin)
    [[ 2.25809786e+03  2.24380103e+03  2.25659060e+03 ... -2.34783682e+03
      -2.30464421e+03 -2.30387946e+03]
     [-1.64627550e+00  2.93583564e-01 -1.06788937e+00 ... -1.64354585e+01
       1.60621841e+01 -1.05506092e+01]
     [-4.15682825e+00  3.08881971e-01 -1.02012415e+00 ...  5.44745701e+00
      -4.71776995e+00  4.39850830e+00]] uK_RJ
    """

    def __init__(
        self,
        nside: int,
        components: Dict[str, SkyComponent],
        version: Optional[str] = None,
    ) -> None:
        """Initializes an instance of the Cosmoglobe Sky Model.

        Parameters
        ----------
        nside
            Healpix resolution of the maps in sky model.
        components
            A list of pre-initialized `SkyComponent`to include in the model.
        version
            The version of the Cosmoglobe Model used in the sky model.
        """

        self.nside = nside
        self.components = components
        self.version = version

        if not all(
            isinstance(component, SkyComponent) for component in components.values()
        ):
            raise ComponentError("all components must be subclasses of SkyComponent")

        if not all(
            self.nside == hp.get_nside(component.amp)
            for component in components.values()
            if not isinstance(component, PointSourceComponent)
        ):
            raise NsideError(
                "all diffuse maps in the sky model needs to be at a common nside"
            )

    @classmethod
    def from_chain(
        cls,
        chain: Union[str, Chain],
        nside: int,
        components: Optional[List[str]] = None,
        model: str = "BeyondPlanck",
        samples: Union[range, int, Literal["all"]] = -1,
        burn_in: Optional[int] = None,
    ) -> SkyModel:
        """Initializes the SkyModel from a Cosmoglobe chain.

        Parameters
        ----------
        chain
            Path to a Cosmoglobe chainfile or a Chain object.
        nside
            Model HEALPIX map resolution parameter.
        components
            List of components to include in the model.
        model
            String representing which sky model to use. Defaults to BeyondPlanck.
        samples
            The sample number for which to extract the model. If the input
            is 'all', then the model will an average of all samples in the chain.
            Defaults to the last sample in the chain.
        burn_in
            Burn in sample for which all previous samples are disregarded.

        Returns
        -------
        sky_model
            Initialized sky model.
        """

        cosmoglobe_model = cosmoglobe_registry.get_model(model)
        initialized_components = get_components_from_chain(
            chain=chain,
            nside=nside,
            components=components,
            model=cosmoglobe_model,
            samples=samples,
            burn_in=burn_in,
        )

        return cls(
            nside=nside,
            components=initialized_components,
            version=cosmoglobe_model.version,
        )

    @classmethod
    def from_hub(cls) -> SkyModel:
        """Initializes the SkyModel from a public cosmoglobe data release."""

        raise NotImplementedError

    @property
    def nside(self) -> int:
        """HEALPIX map resolution of the maps in the Sky Model."""

        return self._nside

    @nside.setter
    def nside(self, value: int) -> None:
        """Validating the nside."""

        try:
            if not hp.isnsideok(value, nest=True):
                raise NsideError("nside needs to be a power of 2")
        except (TypeError, ValueError):
            raise TypeError("nside must be an integer")

        self._nside = int(value)

    def __call__(
        self,
        freqs: Quantity,
        weights: Optional[Quantity] = None,
        *,
        components: Optional[List[str]] = None,
        fwhm: Quantity = DEFAULT_BEAM_FWHM,
        output_unit: str = DEFAULT_OUTPUT_UNIT_STR,
    ) -> Quantity:
        r"""Simulates and returns the full sky model emission.

        Parameters
        ----------
        freqs
            A frequency, or a list of frequencies corresponding to the 
            bandpass weights.
        weights
            Bandpass weights corresponding to the frequencies in `freqs`.
            The units of this quantity must match the units of the related 
            detector, i.e, they must be compatible with K_RJ, K_CMB, or Jy/sr.
            Even if the bandpass weights are already normalized and now have 
            units of 1/Hz, redefine the quantities unites to be that of the 
            detector. If `bandpass` is None and `freqs` is a single frequency, 
            a delta peak is assumed. Defaults to None.
        components
            List of component labels. If None, all components in the sky
            model is included. Defaults to None.
        fwhm
            The full width half max parameter of the Gaussian. Defaults to
            0.0, which indicates no smoothing of output maps.
        output_unit
            The output units of the emission. For instance 'uK_RJ', 'uK_CMB', 
            or 'MJy/sr'. If Kelvin, Rayleigh-Jeans or CMB must be specified.

        Returns
        -------
        emission
            The simulated emission given the Cosmoglobe Sky Model.
        """

        if components is None:
            included_components = list(self.components.values())
        else:
            if not all(component in self.components for component in components):
                raise ValueError("all component must be present in the model")
            included_components = [
                component
                for label, component in self.components.items()
                if label in components
            ]

        diffuse_components: List[Union[DiffuseComponent, LineComponent]] = []
        pointsource_components: List[PointSourceComponent] = []
        for component_class in included_components:
            if isinstance(component_class, (DiffuseComponent, LineComponent)):
                diffuse_components.append(component_class)
            elif isinstance(component_class, PointSourceComponent):
                pointsource_components.append(component_class)

        emission_unit = Unit(output_unit)
        emission = Quantity(np.zeros((3, hp.nside2npix(self.nside))), unit=emission_unit)

        for diffuse_component in diffuse_components:
            component_emission = diffuse_component.simulate_emission(
                freqs=freqs,
                weights=weights,
                nside=self.nside,
                fwhm=fwhm,
                output_unit=emission_unit,
            )
            for IQU, diffuse_emission in enumerate(component_emission):
                emission[IQU] += diffuse_emission

        # We smooth all diffuse components together in a single smoothing operation.
        if fwhm != DEFAULT_BEAM_FWHM:
            emission = Quantity(
                hp.smoothing(emission, fwhm=fwhm.to("rad").value), unit=emission.unit
            )

        # Pointsource emissions are already smoothed during the stage where 
        # each source is mapped to the HEALPIX map.
        for pointsource_component in pointsource_components:
            component_emission = pointsource_component.simulate_emission(
                freqs=freqs,
                weights=weights,
                nside=self.nside,
                fwhm=fwhm,
                output_unit=emission_unit,
            )
            for IQU, pointsource_emission in enumerate(component_emission):
                emission[IQU] += pointsource_emission

        return emission

    def remove_dipole(
        self,
        gal_cut: Quantity = DEFAULT_GAL_CUT,
        return_dipole: bool = False,
    ) -> Optional[Quantity]:
        """Removes the Solar dipole (and monopole), from the CMB sky component.

        Parameters
        ----------
        gal_cut
            Ignores pixels in range [-gal_cut;+gal_cut] when fitting and
            subtracting the dipole.
        return_dipole
            If True, returns the dipole map.

        Returns
        _______
        dipole
            The dipole of the CMB sky component.
        """

        if not any(
            component.label == SkyComponentLabel.CMB
            for component in self.components.values()
        ):
            raise ComponentNotFoundError("CMB component not found in sky model.")

        try:
            gal_cut = gal_cut.to("deg")
        except UnitsError:
            raise UnitsError(
                "gal_cut must be an astropy quantity compatible with 'deg'"
            )

        if return_dipole:
            dipole_subtracted_amp = Quantity(
                hp.remove_dipole(
                    self["cmb"].amp[0],
                    gal_cut=gal_cut.to("deg").value,
                ),
                unit=self["cmb"].amp.unit,
            )
            dipole = self["cmb"].amp[0] - dipole_subtracted_amp
            self["cmb"].amp[0] = dipole_subtracted_amp

            return dipole

        hp.remove_dipole(
            self["cmb"].amp[0], gal_cut=gal_cut.to("deg").value, copy=False
        )

    def __iter__(self) -> Iterator:
        """Returns an iterator over the model components."""

        return iter(list(self.components.values()))

    def __getitem__(self, key: str) -> SkyComponent:
        """Returns a SkyComponent class."""

        try:
            return self.components[key]
        except KeyError:
            raise ComponentNotFoundError(f"component {key} not found in sky model.")

    def __repr__(self) -> str:
        """Representation of the SkyModel and all enabled components."""

        reprs = []
        for label, component in self.components.items():
            component_repr = repr(component) + "\n"
            reprs.append(f"({label}): {component_repr}")

        main_repr = "SkyModel("
        if self.version is not None:
            main_repr += f"\n  version: {self.version}"
        main_repr += f"\n  nside: {self._nside}"
        main_repr += "\n  components( "
        main_repr += "\n    " + "    ".join(reprs)
        main_repr += "  )"
        main_repr += "\n)"

        return main_repr



from typing import Dict, Protocol

import numpy as np
from astropy.units import Quantity

import cosmoglobe.sky._constants as const
from cosmoglobe.sky._units import *


class IntensityDerivative(Protocol):
    """Protocol defining the interface for an intensity derivate."""

    def __call__(self, freqs: Quantity) -> Quantity:
        """Returns the intensity derivative given some frequencies.

        Parameters
        ----------
        freqs
            Frequencies for which to compute the intensity derivative.

        Returns
        -------
            The intensity derivative unit conversion factor.
        """


def bnu_prime_RJ(freqs: Quantity) -> Quantity:
    """Intensitity derivative for K_RJ (dB_nu(T)_dT in the Rayleigh-Jeans limit)."""

    factor = (2 * const.k_B * freqs ** 2) / const.c ** 2
    factor *= Unit("K") / (Unit("K_RJ") * Unit("sr"))

    return factor


def bnu_prime_CMB(freqs: Quantity) -> Quantity:
    """Intensity derivative for K_CMB (dB_nu(T)/dT)."""

    x = (const.h * freqs) / (const.k_B * const.T_0)
    A = 2 * const.k_B * freqs ** 2 / const.c ** 2
    factor = A * x ** 2 * np.exp(x) / np.expm1(x) ** 2
    factor *= Unit("K") / (Unit("K_CMB") * Unit("sr"))

    return factor


def dI_nu_dI_c(freqs: Quantity) -> Quantity:
    """Intensity derivative for MJy/sr (IRAS convention)."""

    return np.mean(freqs) / freqs


INTENSITY_DERIVATIVE_MAPPINGS: Dict[Unit, IntensityDerivative] = {
    Unit("K_RJ"): bnu_prime_RJ,
    Unit("K_CMB"): bnu_prime_CMB,
    Unit("Jy/sr"): dI_nu_dI_c,
}


def get_intensity_derivative(unit: Unit) -> IntensityDerivative:
    """Returns the intensity derivative related to a unit."""

    for intensity_derivative in INTENSITY_DERIVATIVE_MAPPINGS:
        if intensity_derivative.is_equivalent(unit):
            return INTENSITY_DERIVATIVE_MAPPINGS[intensity_derivative]

    raise KeyError("unit does not match any known intensity derivatives.")



class ModelNotFoundError(Exception):
    """Raised when a model is not found in the registry."""


class ComponentNotFoundError(Exception):
    """Raised if a component is missing from the model."""


class CosmoglobeModelError(Exception):
    """Raised if there is an issue with the initialization of a CosmoglobeModel."""


class ComponentError(Exception):
    """Raised if there is an issue with a component in the sky model."""


class NsideError(Exception):
    """Raised if there is an issue with the nside of a map."""



from typing import Tuple

import astropy.constants as const
from astropy.units import Quantity

from cosmoglobe.sky._units import Unit

DEFAULT_BEAM_FWHM: Quantity = 0.0 * Unit("arcmin")
DEFAULT_OUTPUT_UNIT: Unit = Unit("uK_RJ")
DEFAULT_OUTPUT_UNIT_STR = "uK_RJ"
DEFAULT_GAL_CUT: Quantity = 10 * Unit("deg")
SIGNAL_UNITS: Tuple[Unit, Unit, Unit] = (Unit("K_RJ"), Unit("K_CMB"), Unit("Jy/sr"))

h: Quantity = const.h
c: Quantity = const.c
k_B: Quantity = const.k_B
T_0: Quantity = 2.7255 * Unit("K")



from typing import Tuple

import numpy as np
from astropy.units import Quantity, Unit

from cosmoglobe.data import DATA_DIR
from cosmoglobe.sky._base_components import DiffuseComponent
from cosmoglobe.sky._units import cmb_equivalencies
from cosmoglobe.sky.components._labels import SkyComponentLabel

SPDUST2_FILE = DATA_DIR / "spdust2_cnm.dat"

SPDUST2_FREQS, SPDUST2_AMPS = np.loadtxt(SPDUST2_FILE, unpack=True)
SPDUST2_FREQS *= Unit("GHz")
SPDUST2_AMPS *= Unit("Jy/sr")
SPDUST2_TEMPLATE = (SPDUST2_FREQS, SPDUST2_AMPS)


class SpinningDust(DiffuseComponent):
    r"""Class representing the AME component in the Cosmoglobe Sky Model.

    Notes
    -----
    The spinning dust emission is defined using the convention in
    `BeyondPlanck (2020), Section 3.3.4
    <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_{\mathrm{RJ}}^{\mathrm{sd}}(\nu) \propto
        \nu^{-2} \boldsymbol{s}_{0}^{\mathrm{sd}}\left(\nu \cdot
        \frac{30.0\; \mathrm{GHz}}{\nu_{p}}\right)

    where :math:`\nu` is the frequency for which we are simulating the
    sky emission, :math:`\boldsymbol{s}_0^{\mathrm{sd}}` is the `spdust2`
    template, and :math:`\nu_p` the peak frequency.
    """

    label = SkyComponentLabel.AME
    freq_range = (0 * Unit("Hz"), 500 * Unit("GHz"))
    SPINNING_DUST_TEMPLATE: Tuple[Quantity, Quantity] = SPDUST2_TEMPLATE

    def get_freq_scaling(self, freqs: Quantity, freq_peak: Quantity) -> Quantity:
        """See base class."""

        spdust2_freqs = SPDUST2_TEMPLATE[0]
        # We set the spdust2 template amplitudes to the component amplitude units.
        spdust2_amps = SPDUST2_TEMPLATE[1].to(
            self.amp.unit, equivalencies=cmb_equivalencies(spdust2_freqs)
        )

        peak_scale = 30 * Unit("GHz") / freq_peak
        interp = np.interp(
            (freqs * peak_scale).si.value,
            spdust2_freqs.si.value,
            spdust2_amps.decompose().value,
            left=0.0,
            right=0.0,
        )
        interp_ref = np.interp(
            (self.freq_ref * peak_scale).si.value,
            spdust2_freqs.si.value,
            spdust2_amps.decompose().value,
        )

        scaling = interp / interp_ref

        return Quantity(scaling)



import click

from cosmoglobe.fits.click_fits import commands_fits
from cosmoglobe.release.commands_hdf import commands_hdf
from cosmoglobe.h5.click_h5 import commands_h5
from cosmoglobe.plot.click_plotting import commands_plotting

CONTEXT_SETTINGS = dict(max_content_width=100)

@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    pass

cli = click.CommandCollection(sources=[commands_plotting, commands_h5, commands_fits, commands_hdf], context_settings=CONTEXT_SETTINGS)

if __name__ == '__main__':
    cli()
    



from typing import Dict, Protocol

import numpy as np
from astropy.units import Quantity

from cosmoglobe.sky._context_registry import ChainContextRegistry
from cosmoglobe.sky._units import *
from cosmoglobe.sky.components.ame import SpinningDust
from cosmoglobe.sky.components.cmb import CMB, CMB_LOWL
from cosmoglobe.sky.components.dust import ModifiedBlackbody
from cosmoglobe.sky.components.freefree import LinearOpticallyThin
from cosmoglobe.sky.components.radio import AGNPowerLaw
from cosmoglobe.sky.components.synchrotron import PowerLaw


class ChainContext(Protocol):
    """Protocol defining the interface for a chain context.

    Chain context defines additional processing required on the chain items
    before they are ready to be put into the sky model.

    A context needs to manipulate and return the `args` dictionary.

    NOTE: Context functions are executed *AFTER* renaming the units in the
    chain. It is therefore important to use keys to the args dict that match
    the registered naming mappings.
    """

    def __call__(self, args: Dict[str, Quantity]) -> Dict[str, Quantity]:
        ...


def reshape_freq_ref(args: Dict[str, Quantity]) -> Dict[str, Quantity]:
    """Context that re-shapes the `freq_ref` attribute for unpolarized components."""

    if "freq_ref" in args:
        args["freq_ref"] = args["freq_ref"].to("GHz")
        if (amp_dim := args["amp"].shape[0]) == 1:
            args["freq_ref"] = args["freq_ref"][0].reshape((1, 1))
        elif amp_dim == 3:
            args["freq_ref"] = args["freq_ref"].reshape((3, 1))
        else:
            raise ValueError("cannot reshape freq_ref into shape (3,1) or (1,1")

    return args


def radio_specind(args: Dict[str, Quantity]) -> Dict[str, Quantity]:
    """Context that removes all but the first column of alpha.

    We are only interested in the column representinc the power law index
    of the spectral index paramter in the chainfiles.
    """

    args["alpha"] = args["alpha"][0]

    return args


def map_to_scalar(args: Dict[str, Quantity]) -> Dict[str, Quantity]:
    """Extract and returns a scalar.

    Datasets for unsamples quantities the cosmoglobe chains tends to be stored
    in HEALPIX maps. A quantity is considered a scalar if it is constant
    over all axes of the dataset.
    """

    IGNORED_ARGS = ["amp", "freq_ref"]
    for key, value in args.items():
        if key not in IGNORED_ARGS and np.size(value) > 1:
            if np.ndim(value) > 1:
                uniques = [np.unique(col) for col in value]
                all_cols_are_unique = all([len(col) == 1 for col in uniques])
            else:
                uniques = np.unique(value)
                all_cols_are_unique = len(uniques) == 1

            if all_cols_are_unique:
                args[key] = Quantity(uniques, unit=value.unit)

    return args


chain_context_registry = ChainContextRegistry()

chain_context_registry.register_class_context(
    CMB,
    functions=[reshape_freq_ref, map_to_scalar],
    mappings={"freq_ref": "nu_ref"},
    units={"freq_ref": Unit("Hz"), "amp": Unit("uK_CMB")},
)
chain_context_registry.register_class_context(
    CMB_LOWL,
    functions=[reshape_freq_ref, map_to_scalar],
    mappings={"freq_ref": "nu_ref"},
    units={"freq_ref": Unit("Hz"), "amp": Unit("uK_CMB")},
)
chain_context_registry.register_class_context(
    SpinningDust,
    functions=[reshape_freq_ref, map_to_scalar],
    mappings={"freq_ref": "nu_ref", "freq_peak": "nu_p"},
    units={"freq_ref": Unit("Hz"), "amp": Unit("uK_RJ"), "freq_peak": Unit("GHz")},
)
chain_context_registry.register_class_context(
    ModifiedBlackbody,
    functions=[reshape_freq_ref, map_to_scalar],
    mappings={"freq_ref": "nu_ref"},
    units={"freq_ref": Unit("Hz"), "amp": Unit("uK_RJ"), "T": Unit("K")},
)
chain_context_registry.register_class_context(
    AGNPowerLaw,
    functions=[reshape_freq_ref, map_to_scalar, radio_specind],
    mappings={"freq_ref": "nu_ref", "alpha": "specind"},
    units={"freq_ref": Unit("Hz"), "amp": Unit("mJy")},
)
chain_context_registry.register_class_context(
    LinearOpticallyThin,
    functions=[reshape_freq_ref, map_to_scalar],
    mappings={"freq_ref": "nu_ref", "T_e": "Te"},
    units={"freq_ref": Unit("Hz"), "amp": Unit("uK_RJ"), "T_e": Unit("K")},
)
chain_context_registry.register_class_context(
    PowerLaw,
    functions=[reshape_freq_ref, map_to_scalar],
    mappings={"freq_ref": "nu_ref"},
    units={"freq_ref": Unit("Hz"), "amp": Unit("uK_RJ")},
)
